import keyword
import os
import re
import string
import typing
import datetime

from jinja2 import Environment
from jinja2 import FileSystemLoader

import kuber_maker
from kuber_maker import utils

LINE_REGEX = re.compile(r'\n{4,}(?!\s+)')
SINGLE_LINE_REGEX = re.compile(r'\n{3,}(?=\s+)')


TEMPLATE_PATH_ROOT = os.path.realpath(os.path.join(
    os.path.dirname(__file__),
    'templates'
))

ENVIRONMENT = Environment(
    loader=FileSystemLoader(TEMPLATE_PATH_ROOT)
)

_PYTHON_BUILTINS = [
    'abs', 'delattr', 'hash', 'memoryview', 'set',
    'all', 'dict', 'help', 'min', 'setattr',
    'any', 'dir', 'hex', 'next', 'slice',
    'ascii', 'divmod', 'id', 'object', 'sorted',
    'bin', 'enumerate', 'input', 'oct', 'staticmethod',
    'bool', 'eval', 'int', 'open', 'str',
    'breakpoint', 'exec', 'isinstance', 'ord', 'sum',
    'bytearray', 'filter', 'issubclass', 'pow', 'super',
    'bytes', 'float', 'iter', 'print', 'tuple',
    'callable', 'format', 'len', 'property', 'type',
    'chr', 'frozenset', 'list', 'range', 'vars',
    'classmethod', 'getattr', 'locals', 'repr', 'zip',
    'compile', 'globals', 'map', 'reversed', '__import__',
    'complex', 'hasattr', 'max', 'round'
]

IGNORED_ENTITIES = ['Time', 'IntOrString', 'Quantity']


def _has_status(
        package_entities: typing.Dict[str, kuber_maker.Entity]
) -> bool:
    """..."""
    return any([
        'status' in e.properties
        for e in package_entities.values()
    ])


def _collection_status(entity: kuber_maker.Entity) -> dict:
    """..."""
    is_collection = (
        entity.class_name.lower().endswith('list')
        and 'items' in entity.properties
    )

    return {
        'is_collection': is_collection,
        'items': entity.properties.get('items')
    }


def _to_snake_case(value: str) -> str:
    """..."""
    if not value:
        return value

    if value.startswith('API'):
        value = f'Api{value[3:]}'

    value = value.replace('-', '_')
    if keyword.iskeyword(value) or value in _PYTHON_BUILTINS:
        value = f'{value}_'

    output = []
    for index, character in enumerate(value):
        previous = value[index - 1] if index > 0 else ''
        if character.lower() == character:
            output.append(character)
        elif index == 0 or previous.lower() != previous:
            output.append(character.lower())
        else:
            output.append(f'_{character.lower()}')

    return ''.join(output)


def _printable(value: str) -> str:
    """..."""
    return ''.join(list(filter(lambda x: x in string.printable, value)))


ENVIRONMENT.filters['snake_case'] = _to_snake_case
ENVIRONMENT.filters['printable'] = _printable


def render(template_name: str, **kwargs) -> str:
    """..."""
    return ENVIRONMENT.get_template(template_name).render(**kwargs)


def _get_container_entity(
        all_entities: kuber_maker.AllEntities
) -> kuber_maker.Entity:
    """..."""
    version = all_entities.version.replace('.', '_')
    container_package = f'kuber.{version}.core_v1'
    return all_entities.packages[container_package].entities['Container']


def _get_status_entity(
        all_entities: kuber_maker.AllEntities
) -> kuber_maker.Entity:
    """..."""
    version = all_entities.version.replace('.', '_')
    container_package = f'kuber.{version}.meta_v1'
    return all_entities.packages[container_package].entities['Status']


def _containers_location(
        entity: kuber_maker.Entity,
        all_entities: kuber_maker.AllEntities,
        max_depth: int = 10
) -> typing.List[str]:
    """..."""
    if max_depth < 1:
        return []

    if 'containers' in entity.properties:
        return ['containers']

    for key, value in entity.properties.items():
        data_type = value.data_type
        if not data_type.code_import or data_type.api_type == 'array':
            continue

        child_package = data_type.code_import.package
        child_target = data_type.code_import.target

        if child_target in IGNORED_ENTITIES:
            continue

        child = all_entities.packages[child_package].entities[child_target]
        result = _containers_location(child, all_entities, max_depth - 1)
        if result:
            return [utils.to_snake_case(key), *result]

    return []


def _get_kubernetes_api_class_name(entity: kuber_maker.Entity):
    """.."""
    parts = entity.api_version.replace('.k8s.io', '').split('/')
    if entity.api_path.find('rbac') != -1:
        return f'RbacAuthorization{parts[-1].capitalize()}Api'
    result = ''.join([p.capitalize() for p in parts] + ['Api'])
    if result == 'CoordinationV1Api':
        return 'CoordinationApi'
    return result


def render_package(package: str, all_entities: kuber_maker.AllEntities) -> str:
    """..."""
    blocks = []
    entities = all_entities.packages[package].entities
    container_entity = _get_container_entity(all_entities)
    status_entity = _get_status_entity(all_entities)
    include_container = False
    has_resource = False

    for e in entities.values():
        containers_location = '.'.join(_containers_location(e, all_entities))
        include_container = include_container or bool(containers_location)
        has_resource = has_resource or e.is_resource
        kubernetes_api_class_name = _get_kubernetes_api_class_name(e)
        property_ignores = ['apiVersion', 'kind']

        collection = _collection_status(e)
        if collection['is_collection']:
            base_class = 'Collection'
        elif e.is_resource:
            base_class = 'Resource'
        else:
            base_class = 'Definition'
        blocks.append(render(
            'object-class.jinja2',
            entity=e,
            containers=containers_location,
            container_entity=container_entity,
            kubernetes_api_class_name=kubernetes_api_class_name,
            property_ignores=property_ignores,
            collection=collection,
            base_class=base_class
        ))

    data_types = set([
        p.data_type.api_type
        for e in entities.values()
        for p in e.properties.values()
    ])

    extra_imports = []

    if _has_status(entities):
        extra_imports.append(status_entity)
    if include_container:
        extra_imports.append(container_entity)

    # Remove extra imports that are part of the current package to avoid
    # circular import.
    extra_imports = [e for e in extra_imports if package != e.package]

    imports = list(set([
        imp
        for e in (list(entities.values()) + extra_imports)
        for imp in e.imports
        if imp.package != package and imp.target not in IGNORED_ENTITIES
    ]))

    for extra_entity in extra_imports:
        imports.append(kuber_maker.Import(
            target=extra_entity.class_name,
            package=extra_entity.package,
            api_path=extra_entity.api_path,
            reference=extra_entity.api_path
        ))

    import_block = render(
        'module-imports.jinja2',
        imports=sorted(imports),
        data_types=data_types,
        has_resource=has_resource
    )
    blocks.insert(0, import_block)
    lines = '\n'.join(blocks).replace('\r', '').split('\n')
    lines = [l.rstrip() for l in lines]
    contents = '\n'.join(lines)
    contents = re.sub(LINE_REGEX, '\n\n\n', contents)
    contents = f'{contents.rstrip()}\n'
    return re.sub(SINGLE_LINE_REGEX, '\n\n', contents)


def render_root_package(version: str, spec: kuber_maker.ApiSpec):
    """Renders the root version package data."""
    return render(
        'root-package.jinja2',
        spec=spec,
        version=version,
        dt=datetime.datetime.utcnow()
    )



from unittest.mock import MagicMock

import kuber


def test_cli_invocation():
    """
    Should call the callback and then execute a status command action on
    the created (and empty) resource bundle.
    """
    callback = MagicMock()
    kuber.cli(callback, arguments=['status'])
    callback.assert_called_once()


def test_latest_version_stable():
    """Should have a stable latest version."""
    assert not kuber.latest_kube_version.pre_release
    assert 'alpha' not in str(kuber.latest_kube_version)
    assert 'beta' not in str(kuber.latest_kube_version)



import typing
import datetime as _datetime

from kubernetes import client
from kuber import kube_api as _kube_api

from kuber import definitions as _kuber_definitions
from kuber.v1_20.core_v1 import EventSource
from kuber.v1_20.meta_v1 import ListMeta
from kuber.v1_20.meta_v1 import MicroTime
from kuber.v1_20.meta_v1 import ObjectMeta
from kuber.v1_20.core_v1 import ObjectReference


class Event(_kuber_definitions.Resource):
    """
    Event is a report of an event somewhere in the cluster. It
    generally denotes some state change in the system. Events
    have a limited retention time and triggers and messages may
    evolve with time.  Event consumers should not rely on the
    timing of an event with a given Reason reflecting a
    consistent underlying trigger, or the continued existence of
    events with that Reason.  Events should be treated as
    informative, best-effort, supplemental data.
    """

    def __init__(
            self,
            action: str = None,
            deprecated_count: int = None,
            deprecated_first_timestamp: str = None,
            deprecated_last_timestamp: str = None,
            deprecated_source: 'EventSource' = None,
            event_time: 'MicroTime' = None,
            metadata: 'ObjectMeta' = None,
            note: str = None,
            reason: str = None,
            regarding: 'ObjectReference' = None,
            related: 'ObjectReference' = None,
            reporting_controller: str = None,
            reporting_instance: str = None,
            series: 'EventSeries' = None,
            type_: str = None,
    ):
        """Create Event instance."""
        super(Event, self).__init__(
            api_version='events/v1',
            kind='Event'
        )
        self._properties = {
            'action': action if action is not None else '',
            'deprecatedCount': deprecated_count if deprecated_count is not None else None,
            'deprecatedFirstTimestamp': deprecated_first_timestamp if deprecated_first_timestamp is not None else None,
            'deprecatedLastTimestamp': deprecated_last_timestamp if deprecated_last_timestamp is not None else None,
            'deprecatedSource': deprecated_source if deprecated_source is not None else EventSource(),
            'eventTime': event_time if event_time is not None else MicroTime(),
            'metadata': metadata if metadata is not None else ObjectMeta(),
            'note': note if note is not None else '',
            'reason': reason if reason is not None else '',
            'regarding': regarding if regarding is not None else ObjectReference(),
            'related': related if related is not None else ObjectReference(),
            'reportingController': reporting_controller if reporting_controller is not None else '',
            'reportingInstance': reporting_instance if reporting_instance is not None else '',
            'series': series if series is not None else EventSeries(),
            'type': type_ if type_ is not None else '',

        }
        self._types = {
            'action': (str, None),
            'apiVersion': (str, None),
            'deprecatedCount': (int, None),
            'deprecatedFirstTimestamp': (str, None),
            'deprecatedLastTimestamp': (str, None),
            'deprecatedSource': (EventSource, None),
            'eventTime': (MicroTime, None),
            'kind': (str, None),
            'metadata': (ObjectMeta, None),
            'note': (str, None),
            'reason': (str, None),
            'regarding': (ObjectReference, None),
            'related': (ObjectReference, None),
            'reportingController': (str, None),
            'reportingInstance': (str, None),
            'series': (EventSeries, None),
            'type': (str, None),

        }

    @property
    def action(self) -> str:
        """
        action is what action was taken/failed regarding to the
        regarding object. It is machine-readable. This field cannot
        be empty for new Events and it can have at most 128
        characters.
        """
        return self._properties.get('action')

    @action.setter
    def action(self, value: str):
        """
        action is what action was taken/failed regarding to the
        regarding object. It is machine-readable. This field cannot
        be empty for new Events and it can have at most 128
        characters.
        """
        self._properties['action'] = value

    @property
    def deprecated_count(self) -> int:
        """
        deprecatedCount is the deprecated field assuring backward
        compatibility with core.v1 Event type.
        """
        return self._properties.get('deprecatedCount')

    @deprecated_count.setter
    def deprecated_count(self, value: int):
        """
        deprecatedCount is the deprecated field assuring backward
        compatibility with core.v1 Event type.
        """
        self._properties['deprecatedCount'] = value

    @property
    def deprecated_first_timestamp(self) -> str:
        """
        deprecatedFirstTimestamp is the deprecated field assuring
        backward compatibility with core.v1 Event type.
        """
        return self._properties.get('deprecatedFirstTimestamp')

    @deprecated_first_timestamp.setter
    def deprecated_first_timestamp(
            self,
            value: typing.Union[str, _datetime.datetime, _datetime.date]
    ):
        """
        deprecatedFirstTimestamp is the deprecated field assuring
        backward compatibility with core.v1 Event type.
        """
        if isinstance(value, _datetime.datetime):
            value = value.strftime('%Y-%m-%dT%H:%M:%SZ')
        elif isinstance(value, _datetime.date):
            value = value.strftime('%Y-%m-%dT00:00:00Z')
        self._properties['deprecatedFirstTimestamp'] = value

    @property
    def deprecated_last_timestamp(self) -> str:
        """
        deprecatedLastTimestamp is the deprecated field assuring
        backward compatibility with core.v1 Event type.
        """
        return self._properties.get('deprecatedLastTimestamp')

    @deprecated_last_timestamp.setter
    def deprecated_last_timestamp(
            self,
            value: typing.Union[str, _datetime.datetime, _datetime.date]
    ):
        """
        deprecatedLastTimestamp is the deprecated field assuring
        backward compatibility with core.v1 Event type.
        """
        if isinstance(value, _datetime.datetime):
            value = value.strftime('%Y-%m-%dT%H:%M:%SZ')
        elif isinstance(value, _datetime.date):
            value = value.strftime('%Y-%m-%dT00:00:00Z')
        self._properties['deprecatedLastTimestamp'] = value

    @property
    def deprecated_source(self) -> 'EventSource':
        """
        deprecatedSource is the deprecated field assuring backward
        compatibility with core.v1 Event type.
        """
        return self._properties.get('deprecatedSource')

    @deprecated_source.setter
    def deprecated_source(self, value: typing.Union['EventSource', dict]):
        """
        deprecatedSource is the deprecated field assuring backward
        compatibility with core.v1 Event type.
        """
        if isinstance(value, dict):
            value = EventSource().from_dict(value)
        self._properties['deprecatedSource'] = value

    @property
    def event_time(self) -> 'MicroTime':
        """
        eventTime is the time when this Event was first observed. It
        is required.
        """
        return self._properties.get('eventTime')

    @event_time.setter
    def event_time(self, value: typing.Union['MicroTime', dict]):
        """
        eventTime is the time when this Event was first observed. It
        is required.
        """
        if isinstance(value, dict):
            value = MicroTime().from_dict(value)
        self._properties['eventTime'] = value

    @property
    def metadata(self) -> 'ObjectMeta':
        """

        """
        return self._properties.get('metadata')

    @metadata.setter
    def metadata(self, value: typing.Union['ObjectMeta', dict]):
        """

        """
        if isinstance(value, dict):
            value = ObjectMeta().from_dict(value)
        self._properties['metadata'] = value

    @property
    def note(self) -> str:
        """
        note is a human-readable description of the status of this
        operation. Maximal length of the note is 1kB, but libraries
        should be prepared to handle values up to 64kB.
        """
        return self._properties.get('note')

    @note.setter
    def note(self, value: str):
        """
        note is a human-readable description of the status of this
        operation. Maximal length of the note is 1kB, but libraries
        should be prepared to handle values up to 64kB.
        """
        self._properties['note'] = value

    @property
    def reason(self) -> str:
        """
        reason is why the action was taken. It is human-readable.
        This field cannot be empty for new Events and it can have at
        most 128 characters.
        """
        return self._properties.get('reason')

    @reason.setter
    def reason(self, value: str):
        """
        reason is why the action was taken. It is human-readable.
        This field cannot be empty for new Events and it can have at
        most 128 characters.
        """
        self._properties['reason'] = value

    @property
    def regarding(self) -> 'ObjectReference':
        """
        regarding contains the object this Event is about. In most
        cases it's an Object reporting controller implements, e.g.
        ReplicaSetController implements ReplicaSets and this event
        is emitted because it acts on some changes in a ReplicaSet
        object.
        """
        return self._properties.get('regarding')

    @regarding.setter
    def regarding(self, value: typing.Union['ObjectReference', dict]):
        """
        regarding contains the object this Event is about. In most
        cases it's an Object reporting controller implements, e.g.
        ReplicaSetController implements ReplicaSets and this event
        is emitted because it acts on some changes in a ReplicaSet
        object.
        """
        if isinstance(value, dict):
            value = ObjectReference().from_dict(value)
        self._properties['regarding'] = value

    @property
    def related(self) -> 'ObjectReference':
        """
        related is the optional secondary object for more complex
        actions. E.g. when regarding object triggers a creation or
        deletion of related object.
        """
        return self._properties.get('related')

    @related.setter
    def related(self, value: typing.Union['ObjectReference', dict]):
        """
        related is the optional secondary object for more complex
        actions. E.g. when regarding object triggers a creation or
        deletion of related object.
        """
        if isinstance(value, dict):
            value = ObjectReference().from_dict(value)
        self._properties['related'] = value

    @property
    def reporting_controller(self) -> str:
        """
        reportingController is the name of the controller that
        emitted this Event, e.g. `kubernetes.io/kubelet`. This field
        cannot be empty for new Events.
        """
        return self._properties.get('reportingController')

    @reporting_controller.setter
    def reporting_controller(self, value: str):
        """
        reportingController is the name of the controller that
        emitted this Event, e.g. `kubernetes.io/kubelet`. This field
        cannot be empty for new Events.
        """
        self._properties['reportingController'] = value

    @property
    def reporting_instance(self) -> str:
        """
        reportingInstance is the ID of the controller instance, e.g.
        `kubelet-xyzf`. This field cannot be empty for new Events
        and it can have at most 128 characters.
        """
        return self._properties.get('reportingInstance')

    @reporting_instance.setter
    def reporting_instance(self, value: str):
        """
        reportingInstance is the ID of the controller instance, e.g.
        `kubelet-xyzf`. This field cannot be empty for new Events
        and it can have at most 128 characters.
        """
        self._properties['reportingInstance'] = value

    @property
    def series(self) -> 'EventSeries':
        """
        series is data about the Event series this event represents
        or nil if it's a singleton Event.
        """
        return self._properties.get('series')

    @series.setter
    def series(self, value: typing.Union['EventSeries', dict]):
        """
        series is data about the Event series this event represents
        or nil if it's a singleton Event.
        """
        if isinstance(value, dict):
            value = EventSeries().from_dict(value)
        self._properties['series'] = value

    @property
    def type_(self) -> str:
        """
        type is the type of this event (Normal, Warning), new types
        could be added in the future. It is machine-readable. This
        field cannot be empty for new Events.
        """
        return self._properties.get('type')

    @type_.setter
    def type_(self, value: str):
        """
        type is the type of this event (Normal, Warning), new types
        could be added in the future. It is machine-readable. This
        field cannot be empty for new Events.
        """
        self._properties['type'] = value

    def create_resource(self, namespace: 'str' = None):
        """
        Creates the Event in the currently
        configured Kubernetes cluster.
        """
        names = [
            'create_namespaced_event',
            'create_event'
        ]

        _kube_api.execute(
            action='create',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict()}
        )

    def replace_resource(self, namespace: 'str' = None):
        """
        Replaces the Event in the currently
        configured Kubernetes cluster.
        """
        names = [
            'replace_namespaced_event',
            'replace_event'
        ]

        _kube_api.execute(
            action='replace',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict(), 'name': self.metadata.name}
        )

    def patch_resource(self, namespace: 'str' = None):
        """
        Patches the Event in the currently
        configured Kubernetes cluster.
        """
        names = [
            'patch_namespaced_event',
            'patch_event'
        ]

        _kube_api.execute(
            action='patch',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict(), 'name': self.metadata.name}
        )

    def get_resource_status(self, namespace: 'str' = None):
        """This resource does not have a status."""
        pass

    def read_resource(
            self,
            namespace: str = None
    ):
        """
        Reads the Event from the currently configured
        Kubernetes cluster and returns the low-level definition object.
        """
        names = [
            'read_namespaced_event',
            'read_event'
        ]
        return _kube_api.execute(
            action='read',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'name': self.metadata.name}
        )

    def delete_resource(
            self,
            namespace: str = None,
            propagation_policy: str = 'Foreground',
            grace_period_seconds: int = 10
    ):
        """
        Deletes the Event from the currently configured
        Kubernetes cluster.
        """
        names = [
            'delete_namespaced_event',
            'delete_event'
        ]

        body = client.V1DeleteOptions(
            propagation_policy=propagation_policy,
            grace_period_seconds=grace_period_seconds
        )

        _kube_api.execute(
            action='delete',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'name': self.metadata.name, 'body': body}
        )

    @staticmethod
    def get_resource_api(
            api_client: client.ApiClient = None,
            **kwargs
    ) -> 'client.EventsV1Api':
        """
        Returns an instance of the kubernetes API client associated with
        this object.
        """
        if api_client:
            kwargs['apl_client'] = api_client
        return client.EventsV1Api(**kwargs)

    def __enter__(self) -> 'Event':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class EventList(_kuber_definitions.Collection):
    """
    EventList is a list of Event objects.
    """

    def __init__(
            self,
            items: typing.List['Event'] = None,
            metadata: 'ListMeta' = None,
    ):
        """Create EventList instance."""
        super(EventList, self).__init__(
            api_version='events/v1',
            kind='EventList'
        )
        self._properties = {
            'items': items if items is not None else [],
            'metadata': metadata if metadata is not None else ListMeta(),

        }
        self._types = {
            'apiVersion': (str, None),
            'items': (list, Event),
            'kind': (str, None),
            'metadata': (ListMeta, None),

        }

    @property
    def items(self) -> typing.List['Event']:
        """
        items is a list of schema objects.
        """
        return self._properties.get('items')

    @items.setter
    def items(
            self,
            value: typing.Union[typing.List['Event'], typing.List[dict]]
    ):
        """
        items is a list of schema objects.
        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = Event().from_dict(item)
            cleaned.append(item)
        self._properties['items'] = cleaned

    @property
    def metadata(self) -> 'ListMeta':
        """
        Standard list metadata. More info:
        https://git.k8s.io/community/contributors/devel/sig-
        architecture/api-conventions.md#metadata
        """
        return self._properties.get('metadata')

    @metadata.setter
    def metadata(self, value: typing.Union['ListMeta', dict]):
        """
        Standard list metadata. More info:
        https://git.k8s.io/community/contributors/devel/sig-
        architecture/api-conventions.md#metadata
        """
        if isinstance(value, dict):
            value = ListMeta().from_dict(value)
        self._properties['metadata'] = value

    @staticmethod
    def get_resource_api(
            api_client: client.ApiClient = None,
            **kwargs
    ) -> 'client.EventsV1Api':
        """
        Returns an instance of the kubernetes API client associated with
        this object.
        """
        if api_client:
            kwargs['apl_client'] = api_client
        return client.EventsV1Api(**kwargs)

    def __enter__(self) -> 'EventList':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class EventSeries(_kuber_definitions.Definition):
    """
    EventSeries contain information on series of events, i.e.
    thing that was/is happening continuously for some time. How
    often to update the EventSeries is up to the event
    reporters. The default event reporter in "k8s.io/client-
    go/tools/events/event_broadcaster.go" shows how this struct
    is updated on heartbeats and can guide customized reporter
    implementations.
    """

    def __init__(
            self,
            count: int = None,
            last_observed_time: 'MicroTime' = None,
    ):
        """Create EventSeries instance."""
        super(EventSeries, self).__init__(
            api_version='events/v1',
            kind='EventSeries'
        )
        self._properties = {
            'count': count if count is not None else None,
            'lastObservedTime': last_observed_time if last_observed_time is not None else MicroTime(),

        }
        self._types = {
            'count': (int, None),
            'lastObservedTime': (MicroTime, None),

        }

    @property
    def count(self) -> int:
        """
        count is the number of occurrences in this series up to the
        last heartbeat time.
        """
        return self._properties.get('count')

    @count.setter
    def count(self, value: int):
        """
        count is the number of occurrences in this series up to the
        last heartbeat time.
        """
        self._properties['count'] = value

    @property
    def last_observed_time(self) -> 'MicroTime':
        """
        lastObservedTime is the time when last Event from the series
        was seen before last heartbeat.
        """
        return self._properties.get('lastObservedTime')

    @last_observed_time.setter
    def last_observed_time(self, value: typing.Union['MicroTime', dict]):
        """
        lastObservedTime is the time when last Event from the series
        was seen before last heartbeat.
        """
        if isinstance(value, dict):
            value = MicroTime().from_dict(value)
        self._properties['lastObservedTime'] = value

    def __enter__(self) -> 'EventSeries':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False



import os
from unittest.mock import MagicMock
from unittest.mock import patch

import pytest
from pytest import mark

import kuber
from kuber import execution
from kuber import interface

MY_DIRECTORY = os.path.realpath(os.path.dirname(__file__))

RESPONSE_SCENARIOS = [
    execution.ResponseInfo(
        resource=MagicMock(),
        symbol='!!',
        reason='Failure',
        message='Error occurred',
        exception=MagicMock()
    ),
    execution.ResponseInfo(
        resource=MagicMock(),
        symbol='',
        reason='Success',
        message='No error occurred'
    ),
]


@mark.parametrize('response', RESPONSE_SCENARIOS)
@patch('kuber.execution.get_resource_status')
def test_do_status(
        get_resource_status: MagicMock,
        response: execution.ResponseInfo
):
    """Should get statuses of all resources and display them."""
    get_resource_status.return_value = response
    bundle = MagicMock()
    bundle.resources.matching.return_value = [
        MagicMock(),
        MagicMock()
    ]
    action = interface.CommandAction(MagicMock(), [], bundle)
    interface.do_status(action)
    assert get_resource_status.call_count == 2


@mark.parametrize('response', RESPONSE_SCENARIOS)
@patch('kuber.execution.delete_resource')
def test_do_delete(
        delete_resource: MagicMock,
        response: execution.ResponseInfo
):
    """Should execute a delete of all resources in the bundle."""
    delete_resource.return_value = response
    bundle = MagicMock()
    bundle.resources.matching.return_value = [
        MagicMock(),
        MagicMock()
    ]
    action = interface.CommandAction(MagicMock(), [], bundle)
    interface.do_delete(action)
    assert delete_resource.call_count == 2


@mark.parametrize('response', RESPONSE_SCENARIOS)
@patch('kuber.execution.create_resource')
def test_do_create(
        create_resource: MagicMock,
        response: execution.ResponseInfo
):
    """Should execute a creation for all resources in the bundle."""
    create_resource.return_value = response
    bundle = MagicMock()
    bundle.resources.matching.return_value = [
        MagicMock(),
        MagicMock()
    ]
    action = interface.CommandAction(MagicMock(), [], bundle)
    interface.do_create(action)
    assert create_resource.call_count == 2


def test_do_render():
    """Should render all resources in the bundle to the display."""
    bundle = MagicMock()
    resource = MagicMock()
    resource.to_yaml.return_value = ''
    bundle.resources.matching.return_value = [
        resource,
        resource
    ]
    action = interface.CommandAction(MagicMock(), [], bundle)
    interface.do_render(action)


def test_cli():
    """Should execute a render operation without error."""
    interface.ResourceBundleCli(MagicMock())(['render'])


def test_settings_directory():
    """Should load settings from this test directory."""
    def _callback(action: kuber.CommandAction):
        s = action.bundle.settings
        assert s.foo and s.foo == s.spam
        assert s.bar and s.bar == s.ham
        assert s.baz and s.baz == s.eggs

    cb = MagicMock()
    cb.side_effect = _callback
    kuber.cli(cb, arguments=['render', f'--settings={MY_DIRECTORY}'])
    cb.assert_called_once()


def test_settings_files():
    """Should load settings from this test files."""
    def _callback(action: kuber.CommandAction):
        s = action.bundle.settings
        assert s.foo and s.foo == s.spam
        assert s.bar and s.bar == s.ham
        assert s.baz and s.baz == s.eggs

    cb = MagicMock()
    cb.side_effect = _callback

    kuber.cli(cb, arguments=[
        'render',
        f'--settings={os.path.join(MY_DIRECTORY, "settings.yaml")}',
        f'--settings={os.path.join(MY_DIRECTORY, "settings.json")}',
    ])
    cb.assert_called_once()


def test_settings_error():
    """Should load settings from this test files."""
    cb = MagicMock()

    with pytest.raises(ValueError):
        kuber.cli(cb, arguments=['render', '--settings=foo'])

    cb.assert_not_called()


@patch('kuber.execution.get_resource_status')
def test_targeting_all(get_resource_status: MagicMock):
    """Should target all resources if no target is set on the command line."""
    get_resource_status.return_value = RESPONSE_SCENARIOS[-1]
    bundle = kuber.create_bundle()
    bundle.add('v1', 'Namespace', 'foo')
    bundle.add('v1', 'Namespace', 'bar')
    bundle.add('v1', 'Namespace', 'baz')

    bundle.cli(arguments=['status'])
    assert get_resource_status.call_count == 3


@patch('kuber.execution.get_resource_status')
def test_targeting(get_resource_status: MagicMock):
    """Should target the one matching resource specified in the command."""
    get_resource_status.return_value = RESPONSE_SCENARIOS[-1]
    bundle = kuber.create_bundle()
    bundle.add('v1', 'Namespace', 'foo')
    bundle.add('v1', 'Namespace', 'bar')
    bundle.add('v1', 'Namespace', 'baz')

    bundle.cli(arguments=['status', '--target=namespace/foo'])
    assert get_resource_status.call_count == 1



import typing

from kubernetes import client
from kuber import kube_api as _kube_api

from kuber import definitions as _kuber_definitions
from kuber.v1_14.core_v1 import EnvFromSource
from kuber.v1_14.core_v1 import EnvVar
from kuber.v1_14.meta_v1 import LabelSelector
from kuber.v1_14.meta_v1 import ListMeta
from kuber.v1_14.meta_v1 import ObjectMeta
from kuber.v1_14.core_v1 import Volume
from kuber.v1_14.core_v1 import VolumeMount


class PodPreset(_kuber_definitions.Resource):
    """
    PodPreset is a policy resource that defines additional
    runtime requirements for a Pod.
    """

    def __init__(
            self,
            metadata: 'ObjectMeta' = None,
            spec: 'PodPresetSpec' = None,
    ):
        """Create PodPreset instance."""
        super(PodPreset, self).__init__(
            api_version='settings/v1alpha1',
            kind='PodPreset'
        )
        self._properties = {
            'metadata': metadata if metadata is not None else ObjectMeta(),
            'spec': spec if spec is not None else PodPresetSpec(),

        }
        self._types = {
            'apiVersion': (str, None),
            'kind': (str, None),
            'metadata': (ObjectMeta, None),
            'spec': (PodPresetSpec, None),

        }

    @property
    def metadata(self) -> 'ObjectMeta':
        """

        """
        return self._properties.get('metadata')

    @metadata.setter
    def metadata(self, value: typing.Union['ObjectMeta', dict]):
        """

        """
        if isinstance(value, dict):
            value = ObjectMeta().from_dict(value)
        self._properties['metadata'] = value

    @property
    def spec(self) -> 'PodPresetSpec':
        """

        """
        return self._properties.get('spec')

    @spec.setter
    def spec(self, value: typing.Union['PodPresetSpec', dict]):
        """

        """
        if isinstance(value, dict):
            value = PodPresetSpec().from_dict(value)
        self._properties['spec'] = value

    def create_resource(self, namespace: 'str' = None):
        """
        Creates the PodPreset in the currently
        configured Kubernetes cluster.
        """
        names = [
            'create_namespaced_pod_preset',
            'create_pod_preset'
        ]

        _kube_api.execute(
            action='create',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict()}
        )

    def replace_resource(self, namespace: 'str' = None):
        """
        Replaces the PodPreset in the currently
        configured Kubernetes cluster.
        """
        names = [
            'replace_namespaced_pod_preset',
            'replace_pod_preset'
        ]

        _kube_api.execute(
            action='replace',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict(), 'name': self.metadata.name}
        )

    def patch_resource(self, namespace: 'str' = None):
        """
        Patches the PodPreset in the currently
        configured Kubernetes cluster.
        """
        names = [
            'patch_namespaced_pod_preset',
            'patch_pod_preset'
        ]

        _kube_api.execute(
            action='patch',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict(), 'name': self.metadata.name}
        )

    def get_resource_status(self, namespace: 'str' = None):
        """This resource does not have a status."""
        pass

    def read_resource(
            self,
            namespace: str = None
    ):
        """
        Reads the PodPreset from the currently configured
        Kubernetes cluster and returns the low-level definition object.
        """
        names = [
            'read_namespaced_pod_preset',
            'read_pod_preset'
        ]
        return _kube_api.execute(
            action='read',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'name': self.metadata.name}
        )

    def delete_resource(
            self,
            namespace: str = None,
            propagation_policy: str = 'Foreground',
            grace_period_seconds: int = 10
    ):
        """
        Deletes the PodPreset from the currently configured
        Kubernetes cluster.
        """
        names = [
            'delete_namespaced_pod_preset',
            'delete_pod_preset'
        ]

        body = client.V1DeleteOptions(
            propagation_policy=propagation_policy,
            grace_period_seconds=grace_period_seconds
        )

        _kube_api.execute(
            action='delete',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'name': self.metadata.name, 'body': body}
        )

    @staticmethod
    def get_resource_api(
            api_client: client.ApiClient = None,
            **kwargs
    ) -> 'client.SettingsV1alpha1Api':
        """
        Returns an instance of the kubernetes API client associated with
        this object.
        """
        if api_client:
            kwargs['apl_client'] = api_client
        return client.SettingsV1alpha1Api(**kwargs)

    def __enter__(self) -> 'PodPreset':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class PodPresetList(_kuber_definitions.Collection):
    """
    PodPresetList is a list of PodPreset objects.
    """

    def __init__(
            self,
            items: typing.List['PodPreset'] = None,
            metadata: 'ListMeta' = None,
    ):
        """Create PodPresetList instance."""
        super(PodPresetList, self).__init__(
            api_version='settings/v1alpha1',
            kind='PodPresetList'
        )
        self._properties = {
            'items': items if items is not None else [],
            'metadata': metadata if metadata is not None else ListMeta(),

        }
        self._types = {
            'apiVersion': (str, None),
            'items': (list, PodPreset),
            'kind': (str, None),
            'metadata': (ListMeta, None),

        }

    @property
    def items(self) -> typing.List['PodPreset']:
        """
        Items is a list of schema objects.
        """
        return self._properties.get('items')

    @items.setter
    def items(
            self,
            value: typing.Union[typing.List['PodPreset'], typing.List[dict]]
    ):
        """
        Items is a list of schema objects.
        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = PodPreset().from_dict(item)
            cleaned.append(item)
        self._properties['items'] = cleaned

    @property
    def metadata(self) -> 'ListMeta':
        """
        Standard list metadata. More info:
        https://git.k8s.io/community/contributors/devel/sig-
        architecture/api-conventions.md#metadata
        """
        return self._properties.get('metadata')

    @metadata.setter
    def metadata(self, value: typing.Union['ListMeta', dict]):
        """
        Standard list metadata. More info:
        https://git.k8s.io/community/contributors/devel/sig-
        architecture/api-conventions.md#metadata
        """
        if isinstance(value, dict):
            value = ListMeta().from_dict(value)
        self._properties['metadata'] = value

    @staticmethod
    def get_resource_api(
            api_client: client.ApiClient = None,
            **kwargs
    ) -> 'client.SettingsV1alpha1Api':
        """
        Returns an instance of the kubernetes API client associated with
        this object.
        """
        if api_client:
            kwargs['apl_client'] = api_client
        return client.SettingsV1alpha1Api(**kwargs)

    def __enter__(self) -> 'PodPresetList':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class PodPresetSpec(_kuber_definitions.Definition):
    """
    PodPresetSpec is a description of a pod preset.
    """

    def __init__(
            self,
            env: typing.List['EnvVar'] = None,
            env_from: typing.List['EnvFromSource'] = None,
            selector: 'LabelSelector' = None,
            volume_mounts: typing.List['VolumeMount'] = None,
            volumes: typing.List['Volume'] = None,
    ):
        """Create PodPresetSpec instance."""
        super(PodPresetSpec, self).__init__(
            api_version='settings/v1alpha1',
            kind='PodPresetSpec'
        )
        self._properties = {
            'env': env if env is not None else [],
            'envFrom': env_from if env_from is not None else [],
            'selector': selector if selector is not None else LabelSelector(),
            'volumeMounts': volume_mounts if volume_mounts is not None else [],
            'volumes': volumes if volumes is not None else [],

        }
        self._types = {
            'env': (list, EnvVar),
            'envFrom': (list, EnvFromSource),
            'selector': (LabelSelector, None),
            'volumeMounts': (list, VolumeMount),
            'volumes': (list, Volume),

        }

    @property
    def env(self) -> typing.List['EnvVar']:
        """
        Env defines the collection of EnvVar to inject into
        containers.
        """
        return self._properties.get('env')

    @env.setter
    def env(
            self,
            value: typing.Union[typing.List['EnvVar'], typing.List[dict]]
    ):
        """
        Env defines the collection of EnvVar to inject into
        containers.
        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = EnvVar().from_dict(item)
            cleaned.append(item)
        self._properties['env'] = cleaned

    @property
    def env_from(self) -> typing.List['EnvFromSource']:
        """
        EnvFrom defines the collection of EnvFromSource to inject
        into containers.
        """
        return self._properties.get('envFrom')

    @env_from.setter
    def env_from(
            self,
            value: typing.Union[typing.List['EnvFromSource'], typing.List[dict]]
    ):
        """
        EnvFrom defines the collection of EnvFromSource to inject
        into containers.
        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = EnvFromSource().from_dict(item)
            cleaned.append(item)
        self._properties['envFrom'] = cleaned

    @property
    def selector(self) -> 'LabelSelector':
        """
        Selector is a label query over a set of resources, in this
        case pods. Required.
        """
        return self._properties.get('selector')

    @selector.setter
    def selector(self, value: typing.Union['LabelSelector', dict]):
        """
        Selector is a label query over a set of resources, in this
        case pods. Required.
        """
        if isinstance(value, dict):
            value = LabelSelector().from_dict(value)
        self._properties['selector'] = value

    @property
    def volume_mounts(self) -> typing.List['VolumeMount']:
        """
        VolumeMounts defines the collection of VolumeMount to inject
        into containers.
        """
        return self._properties.get('volumeMounts')

    @volume_mounts.setter
    def volume_mounts(
            self,
            value: typing.Union[typing.List['VolumeMount'], typing.List[dict]]
    ):
        """
        VolumeMounts defines the collection of VolumeMount to inject
        into containers.
        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = VolumeMount().from_dict(item)
            cleaned.append(item)
        self._properties['volumeMounts'] = cleaned

    @property
    def volumes(self) -> typing.List['Volume']:
        """
        Volumes defines the collection of Volume to inject into the
        pod.
        """
        return self._properties.get('volumes')

    @volumes.setter
    def volumes(
            self,
            value: typing.Union[typing.List['Volume'], typing.List[dict]]
    ):
        """
        Volumes defines the collection of Volume to inject into the
        pod.
        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = Volume().from_dict(item)
            cleaned.append(item)
        self._properties['volumes'] = cleaned

    def __enter__(self) -> 'PodPresetSpec':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False



from unittest.mock import MagicMock
from unittest.mock import patch

import kuber
from kuber import kube_api
from kuber.tests import utils as test_utils


@patch('kubernetes.config.load_incluster_config')
@patch('kubernetes.config.load_kube_config')
def test_load_access_config(
        load_kube_config: MagicMock,
        load_incluster_config: MagicMock
):
    """Should load cluster configs."""
    kube_api.load_access_config()
    load_kube_config.assert_called_once()
    load_incluster_config.assert_not_called()


@patch('kubernetes.config.load_incluster_config')
@patch('kubernetes.config.load_kube_config')
def test_load_access_config_in_cluster(
        load_kube_config: MagicMock,
        load_incluster_config: MagicMock
):
    """Should load in-cluster configs."""
    kube_api.load_access_config(in_cluster=True)
    load_kube_config.assert_not_called()
    load_incluster_config.assert_called_once()


@patch('kubernetes.client.VersionApi.get_code')
def test_get_version_from_cluster(get_code: MagicMock):
    """Should return the expected "latest" version."""
    latest = kuber.latest_kube_version
    cluster_version = MagicMock()
    cluster_version.major = latest.major
    cluster_version.minor = latest.minor
    get_code.return_value = cluster_version
    result = kube_api.get_version_from_cluster()
    assert result.version == latest.version


@patch('kubernetes.client.VersionApi.get_code')
def test_get_version_from_cluster_error(get_code: MagicMock):
    """Should return the expected "latest" version as the fallback on error."""
    latest = kuber.latest_kube_version
    get_code.side_effect = test_utils.MockApiException(False)
    result = kube_api.get_version_from_cluster(latest)
    assert result.version == latest.version


@patch('kubernetes.client.VersionApi.get_code')
def test_get_version_from_cluster_error_label(get_code: MagicMock):
    """Should return the expected "latest" version as the fallback on error."""
    latest = kuber.latest_kube_version
    get_code.side_effect = test_utils.MockApiException(False)
    result = kube_api.get_version_from_cluster(latest.label)
    assert result.version == latest.version


@patch('kubernetes.client.VersionApi.get_code')
def test_get_version_from_cluster_error_no_fallback(get_code: MagicMock):
    """Should return the expected "latest" version as the fallback on error."""
    get_code.side_effect = test_utils.MockApiException(False)
    result = kube_api.get_version_from_cluster()
    assert isinstance(result, kuber.KubernetesVersion)



import os
import kuber

import pytest

MY_DIRECTORY = os.path.realpath(os.path.dirname(__file__))


def test_resource_arrays():
    """Should access resources using dynamic arrays."""
    bundle = kuber.from_directory(MY_DIRECTORY)
    expected = bundle.get(name='first-a', kind='ConfigMap')
    assert expected == bundle.resources.config_map.first_a
    assert expected in bundle.resources.config_map
    assert expected == bundle.resources.config_map['first-a']
    assert expected == bundle.resources['config_map']['first-a']
    assert expected == bundle.resources['config-map']['first-a']
    assert expected == bundle.resources['ConfigMap']['first-a']


def test_resource_arrays_multi():
    """Should access resources using dynamic arrays."""
    bundle = kuber.from_directory(MY_DIRECTORY)
    expected = bundle.get(
        name='foo-pod',
        kind='Pod',
        namespace='first'
    )
    assert len(bundle.resources.pod) >= 3
    assert len(bundle.resources.pod.foo_pod) == 3
    assert expected in bundle.resources.pod
    assert expected == bundle.resources.pod.foo_pod[0]
    assert expected == bundle.resources.within('first').pod.foo_pod
    assert expected == bundle.resources.within('first').pod['foo-pod']


def test_resource_arrays_missing():
    """Should raise ValueError if resource does not exist."""
    bundle = kuber.from_directory(MY_DIRECTORY)

    with pytest.raises(ValueError):
        print(bundle.resources.pod.this_pod_does_not_exist)

    assert bundle.resources.pod[0] is not None


def test_resource_arrays_indexing():
    """Should retrieve the first loaded pod."""
    bundle = kuber.from_directory(MY_DIRECTORY)
    pods = bundle.get_many(kind='Pod')
    assert pods == bundle.resources.pod.to_list()
    assert pods[0] == bundle.resources.pod[0]


def test_resource_arrays_to_list():
    """Should create a list matching the contents of the array."""
    bundle = kuber.from_directory(MY_DIRECTORY)
    all_resources = bundle.resources.to_list()
    for i, r in enumerate(all_resources):
        assert bundle.resources[i] == all_resources[i]



import os

import kuber
from kuber.latest import rbac_v1

MY_DIRECTORY = os.path.realpath(os.path.dirname(__file__))


def test_empty_api_group():
    """Should include the empty apiGroup value."""
    path = os.path.join(MY_DIRECTORY, 'empty-api-group.yaml')
    cluster_role: rbac_v1.ClusterRole = kuber.from_yaml_file(path)
    assert cluster_role.rules[0].api_groups is not None
    assert 'apiGroups' in cluster_role.to_dict()['rules'][0]



import typing
from collections import defaultdict

import kuber_maker

TYPES = {
    'string': 'str',
    'array': 'list',
    'object': 'dict',
    'integer': 'int',
    'boolean': 'bool',
    'number': 'float',
    'Time': 'str',
    # IntOrStrings must be integer values because of OpenApi v2
    # validation limitations. See issue for details:
    # https://github.com/kubernetes-client/python/issues/322
    'IntOrString': 'int',
    'Quantity': 'str',
    None: None
}

TYPE_HINTS = {
    'IntOrString': 'typing.Union[str, int, None]',
    'Quantity': 'typing.Union[str, int, None]'
}

TYPE_CONSTRUCTORS = {
    'string': '\'\'',
    'array': '[]',
    'object': '{}',
    'integer': 'None',
    'boolean': 'None',
    'number': 'None',
    'Time': 'None',
    'IntOrString': 'None',
    'Quantity': 'None'
}

QUALIFIED_MAPPINGS = {
    'rbac': 'rbac.authorization.k8s.io',
    'apiregistration': 'apiregistration.k8s.io'
}

PATHS_TO_SKIP = (
    'io.k8s.kubernetes.',
    'io.k8s.apimachinery.pkg.api.resource.',
    'io.k8s.apimachinery.pkg.util.',
)


def parse_definitions(
        version: str,
        definitions: dict
) -> kuber_maker.AllEntities:
    """..."""
    entities = defaultdict(dict)
    for api_path, definition in definitions.items():
        if api_path.startswith(PATHS_TO_SKIP):
            continue
        entity = _create_entity(version, api_path, definition)
        entities[entity.package][entity.class_name] = entity
    return kuber_maker.AllEntities(
        version=version,
        packages={
            k: kuber_maker.EntityPackage(k, v)
            for k, v in entities.items()
        }
    )


def parse_property(
        version: str,
        name: str,
        definition: dict,
        entity_class_name: str
) -> kuber_maker.Property:
    """..."""
    definition = definition.copy()
    return kuber_maker.Property(
        name=name,
        data_type=_get_data_type(version, name, definition, entity_class_name),
        description=definition.get('description', ''),
        source=definition
    )


def _get_qualified_api_version(package: str) -> str:
    """..."""
    # Get the api version parts like (app, v1) from the package.
    api_version_parts = package.rsplit('.', 1)[-1].split('_')
    api_version_parts[0] = QUALIFIED_MAPPINGS.get(
        api_version_parts[0],
        api_version_parts[0]
    )

    return '/'.join(api_version_parts)


def _create_entity(
        version: str,
        api_path: str,
        definition: dict
) -> kuber_maker.Entity:
    """..."""
    class_name = api_path.rsplit('.', 1)[-1]
    package = kuber_maker.to_kuber_package(version, api_path)
    path = kuber_maker.to_kuber_path(version, api_path)

    properties: typing.Dict[str, kuber_maker.Property] = {
        name: parse_property(version, name, value, class_name)
        for name, value in definition.get('properties', {}).items()
        if not name.startswith('$')
    }
    imports = sorted(list(set([
        p.data_type.code_import
        for p in properties.values()
        if p.data_type.code_import
    ])))
    is_resource = (
        ('apiVersion' in properties)
        and ('kind' in properties)
        and ('metadata' in properties)
        and (properties['metadata'].data_type.api_type == 'ObjectMeta')
    )

    return kuber_maker.Entity(
        api_path=api_path,
        class_name=class_name,
        api_version=_get_qualified_api_version(package),
        package=package,
        code_path=f'{path}.py',
        description=definition.get('description', ''),
        properties=properties,
        data_type=TYPES[definition.get('type', 'object')],
        source=definition,
        imports=imports,
        is_resource=is_resource
    )


def _get_data_type(
        version: str,
        name: str,
        property_definition: dict,
        entity_class_name: str
) -> kuber_maker.DataType:
    """..."""
    setter_type_hint = None
    reference = property_definition.get('$ref')
    reference_type = (reference or '').rsplit('.', 1)[-1]
    api_type = property_definition.get('type') or reference_type

    basic_type = TYPES.get(api_type)
    python_type = basic_type or reference_type
    constructor = TYPE_CONSTRUCTORS.get(api_type) or f'{reference_type}()'

    if api_type == 'array':
        items = property_definition.get('items') or {}
        reference = items.get('$ref')
        basic_item_type = TYPES[items.get('type')]
        reference_item_type = (reference or '').rsplit('.', 1)[-1]
        item_type = basic_item_type or reference_item_type
        sub_type_hint = item_type if basic_item_type else f'\'{item_type}\''
        type_hint = f'typing.List[{sub_type_hint}]'
    else:
        item_type = None
        type_hint = TYPE_HINTS.get(
            api_type,
            basic_type or f'\'{reference_type}\''
        )

    code_import = (
        kuber_maker.import_from_reference(version, reference)
        if reference else
        None
    )
    if code_import:
        # When the type is a kuber definition, allow for a dict type as well
        # that can be deserialized into the strict object type instead.
        setter_type_hint = f'typing.Union[{type_hint}, dict]'

    if reference_type == entity_class_name:
        # There are places in the API where recursive relationships exist
        # as properties have the same type as the object in which they are
        # defined. To prevent that, we set the values of those properties to
        # None by default and they have to be added manually instead.
        type_hint = f'typing.Optional[\'{reference_type}\']'
        setter_type_hint = f'typing.Union[\'{reference_type}\', dict, None]'
        constructor = 'None'

    return kuber_maker.DataType(
        api_type=api_type,
        python_type=python_type,
        constructor=constructor,
        type_hint=type_hint,
        code_import=code_import,
        item_type=item_type,
        setter_type_hint=setter_type_hint
    )



import datetime as _datetime

from kuber import versioning

# Information about the kubernetes API library version associated with this
# package and kuber metadata around its creation and labeling within the
# kuber library.
KUBERNETES_VERSION = versioning.KubernetesVersion(
    label='v1.17',
    version='v1.17.14',
    major='1',
    minor='17',
    patch='14',
    pre_release='',
    build='',
    created_at=_datetime.datetime(2020, 11, 23),
    commit_sha='f238f5142728be4033c37aa0ad69bf806090beae'
)





def to_snake_case(value: str) -> str:
    if not value:
        return value

    characters = [value[0].lower()]
    for previous_index, character in enumerate(value[1:]):
        previous_character = value[previous_index]
        was_lowercase = previous_character.lower() == previous_character
        is_lowercase = character.lower() == character
        characters.append(
            f'_{character.lower()}'
            if was_lowercase and not is_lowercase else
            character.lower()
        )

    return ''.join(characters)



import os

import kuber
from kuber.latest import core_v1

MY_DIRECTORY = os.path.realpath(os.path.dirname(__file__))


def test_target_port():
    """
    Should return target port as integer because the OpenApi v2 spec used
    by the Python Kubernetes client doesn't support multi-type validation and
    so strings cannot be used at this time as an alternative. For more details
    see the issue:
        - https://github.com/kubernetes-client/python/issues/322
    """
    path = os.path.join(MY_DIRECTORY, 'target-port.yaml')
    service: core_v1.Service = kuber.from_yaml_file(path)
    assert isinstance(service.spec.ports[0].target_port, int)



import os
from unittest.mock import MagicMock
from unittest.mock import patch

import pytest

import kuber
from kuber import latest
from kuber.management import creation

MY_DIRECTORY = os.path.realpath(os.path.dirname(__file__))


def test_initialization():
    """Should initialize bundle as expected."""
    bundle = kuber.create_bundle(
        kubernetes_version='pre',
        bundle_name='foo'
    )
    assert bundle.kubernetes_version.label == 'pre'
    assert bundle.name == 'foo'
    assert bundle.cli is not None


def test_version():
    """Should initialize bundle as expected."""
    bundle = kuber.create_bundle(
        kubernetes_version=latest.KUBERNETES_VERSION,
        bundle_name='foo'
    )
    assert bundle.kubernetes_version.label == 'latest'
    assert bundle.name == 'foo'


def test_new_resource():
    """Should create a new resource from specified arguments."""
    config_map = kuber.new_resource(
        api_version='core/v1',
        kind='ConfigMap',
        kubernetes_version=kuber.latest_kube_version.label.strip('v')
    )
    assert config_map is not None
    assert config_map.kind == 'ConfigMap'


def test_from_dict():
    """Should create a resource from a configuration dictionary."""
    version = kuber.latest_kube_version
    config_map = creation.from_dict(
        resource_definition={
            'apiVersion': 'core/v1',
            'kind': 'ConfigMap',
            'metadata': {'name': 'foo'}
        },
        kubernetes_version=f'{version.major}.{version.minor}'
    )
    assert config_map is not None
    assert config_map.kind == 'ConfigMap'


def test_add_resource():
    """Should create a new resource from specified arguments."""
    bundle = kuber.ResourceBundle()
    bundle.add(
        api_version='core/v1',
        kind='ConfigMap',
        name='foo'
    )
    assert len(bundle.resources) == 1
    assert bundle.resources[0].kind == 'ConfigMap'


def test_add_file():
    """Should add the specified YAML file to the bundle."""
    bundle = kuber.ResourceBundle()
    bundle.add_file(os.path.join(MY_DIRECTORY, 'test-get.yaml'))
    assert len(bundle.resources) > 0


def test_add_file_invalid():
    """Should raise an IOError for unknown file types."""
    bundle = kuber.ResourceBundle()
    with pytest.raises(IOError):
        bundle.add_file(os.path.join(MY_DIRECTORY, 'name.foo'))


def test_add_from_yaml():
    """Should add the specified YAML string to the bundle."""
    path = os.path.join(MY_DIRECTORY, 'test-get.yaml')
    with open(path) as f:
        contents = f.read()

    bundle = kuber.ResourceBundle()
    bundle.add_from_yaml(contents)
    assert len(bundle.resources) > 0


def test_remove():
    """Should return the correct resource type."""
    bundle = kuber.from_file(os.path.join(MY_DIRECTORY, 'test-get.yaml'))
    count = len(bundle.resources)
    resource = bundle.resources[1]
    bundle.remove(resource)
    assert len(bundle.resources) == count - 1
    assert resource not in bundle.resources


def test_pop():
    """Should remove the resource from the bundle."""
    bundle = kuber.from_file(os.path.join(MY_DIRECTORY, 'test-get.yaml'))
    resource = bundle.resources[1]
    popped_resource = bundle.pop(
        name=resource.metadata.name,
        kind=resource.kind,
        **resource.metadata.labels
    )
    assert resource == popped_resource


def test_unshift():
    """
    Should place the resource at the beginning fo the resources list.
    """
    bundle = kuber.from_file(os.path.join(MY_DIRECTORY, 'test-get.yaml'))
    resource = bundle.resources[1]
    bundle.remove(resource)
    bundle.unshift(resource)
    assert bundle.resources[0] == resource


def test_load_directory():
    """Should load all resources in the directory."""
    bundle = kuber.from_directory(MY_DIRECTORY)
    resource = bundle.get(name='loaded-from-json')
    assert resource is not None
    assert len(bundle.resources) > 1


def test_render_json():
    """Should render bundle to JSON."""
    version = latest.KUBERNETES_VERSION
    bundle = kuber.from_directory(
        directory=MY_DIRECTORY,
        kubernetes_version=f'{version.major}.{version.minor}'
    )
    results = bundle.render_json()
    assert len(bundle.resources) == len(results)


def test_render_yaml_bundle():
    """Should render bundle to a single YAML file."""
    version = latest.KUBERNETES_VERSION
    bundle = kuber.from_directory(
        directory=MY_DIRECTORY,
        kubernetes_version=f'{version.major}.{version.minor}'
    )
    assert bundle.render_yaml_bundle()


@patch('kuber.execution.create_resource')
def test_create(create_resource: MagicMock):
    """Should execute create on bundle resources."""
    bundle = kuber.from_file(os.path.join(MY_DIRECTORY, 'test-get.yaml'))
    bundle.create()
    assert create_resource.call_count == len(bundle.resources)


@patch('kuber.execution.replace_resource')
def test_replace(replace_resource: MagicMock):
    """Should execute replace on bundle resources."""
    bundle = kuber.from_file(os.path.join(MY_DIRECTORY, 'test-get.yaml'))
    bundle.replace()
    assert replace_resource.call_count == len(bundle.resources)


@patch('kuber.execution.get_resource_status')
def test_statuses(get_resource_status: MagicMock):
    """Should execute statuses on bundle resources."""
    bundle = kuber.from_file(os.path.join(MY_DIRECTORY, 'test-get.yaml'))
    bundle.statuses()
    assert get_resource_status.call_count == len(bundle.resources)


@patch('kuber.execution.delete_resource')
def test_delete(delete_resource: MagicMock):
    """Should execute delete on bundle resources."""
    bundle = kuber.from_file(os.path.join(MY_DIRECTORY, 'test-get.yaml'))
    bundle.delete()
    assert delete_resource.call_count == len(bundle.resources)



import datetime as _datetime

from kuber import versioning

# Information about the kubernetes API library version associated with this
# package and kuber metadata around its creation and labeling within the
# kuber library.
KUBERNETES_VERSION = versioning.KubernetesVersion(
    label='v1.20',
    version='v1.20.0-beta.2',
    major='1',
    minor='20',
    patch='0',
    pre_release='beta.2',
    build='',
    created_at=_datetime.datetime(2020, 11, 23),
    commit_sha='3af376d3ad5009fa82c94c1ef1fed17dc1f8c29a'
)



import kuber


def test_settings():
    """
    Should properly assign a setting that is accessible through all
    possible avenues.
    """
    s = kuber.create_bundle().settings
    s.add(a=42)
    assert s.fetch('a') == 42
    assert s.grab('a')[0] == 42
    assert s['a'] == 42
    assert s.a == 42
    assert s.to_dict() == {'a': 42}
    assert s.to_yaml().find('42') > 0
    assert list(s.values()) == [42]
    assert list(s.keys()) == ['a']
    assert list(s.items()) == [('a', 42)]



import os
import kuber

MY_DIRECTORY = os.path.realpath(os.path.dirname(__file__))


def test_get_deployment():
    """Should return the correct resource type."""
    bundle = kuber.from_file(os.path.join(MY_DIRECTORY, 'test-get.yaml'))
    resource = bundle.get(name='foo', kind='Deployment')
    assert resource.kind == 'Deployment'


def test_get_config_map():
    """Should return the correct resource type."""
    bundle = kuber.from_file(os.path.join(MY_DIRECTORY, 'test-get.yaml'))
    resource = bundle.get(name='foo', kind='ConfigMap')
    assert resource.kind == 'ConfigMap'


def test_get_job():
    """Should return the correct resource type."""
    bundle = kuber.from_file(os.path.join(MY_DIRECTORY, 'test-get.yaml'))
    resource = bundle.get(name='foo', kind='Job')
    assert resource.kind == 'Job'


def test_get_job_label():
    """Should return the correct resource type."""
    bundle = kuber.from_file(os.path.join(MY_DIRECTORY, 'test-get.yaml'))
    resource = bundle.get(kind='Job', app='bar')
    assert resource.kind == 'Job'
    assert resource.metadata.labels['app'] == 'bar'


def test_get_none():
    """Should not find a match that doesn't exist."""
    bundle = kuber.from_file(os.path.join(MY_DIRECTORY, 'test-get.yaml'))
    resource = bundle.get(app='baz')
    assert resource is None


def test_get_many():
    """Should return all matching resources"""
    bundle = kuber.from_file(os.path.join(MY_DIRECTORY, 'test-get.yaml'))
    resources = bundle.get_many(kind='Job')
    assert len(resources) == 2
    assert set([r.kind for r in resources]) == {'Job'}



import typing

from kubernetes import client
from kuber import kube_api as _kube_api

from kuber import definitions as _kuber_definitions
from kuber.v1_12.meta_v1 import ListMeta
from kuber.v1_12.meta_v1 import ObjectMeta
from kuber.v1_12.meta_v1 import Status
from kuber.v1_12.meta_v1 import StatusDetails


class TokenReview(_kuber_definitions.Resource):
    """
    TokenReview attempts to authenticate a token to a known
    user. Note: TokenReview requests may be cached by the
    webhook token authenticator plugin in the kube-apiserver.
    """

    def __init__(
            self,
            metadata: 'ObjectMeta' = None,
            spec: 'TokenReviewSpec' = None,
            status: 'TokenReviewStatus' = None,
    ):
        """Create TokenReview instance."""
        super(TokenReview, self).__init__(
            api_version='authentication/v1beta1',
            kind='TokenReview'
        )
        self._properties = {
            'metadata': metadata if metadata is not None else ObjectMeta(),
            'spec': spec if spec is not None else TokenReviewSpec(),
            'status': status if status is not None else TokenReviewStatus(),

        }
        self._types = {
            'apiVersion': (str, None),
            'kind': (str, None),
            'metadata': (ObjectMeta, None),
            'spec': (TokenReviewSpec, None),
            'status': (TokenReviewStatus, None),

        }

    @property
    def metadata(self) -> 'ObjectMeta':
        """

        """
        return self._properties.get('metadata')

    @metadata.setter
    def metadata(self, value: typing.Union['ObjectMeta', dict]):
        """

        """
        if isinstance(value, dict):
            value = ObjectMeta().from_dict(value)
        self._properties['metadata'] = value

    @property
    def spec(self) -> 'TokenReviewSpec':
        """
        Spec holds information about the request being evaluated
        """
        return self._properties.get('spec')

    @spec.setter
    def spec(self, value: typing.Union['TokenReviewSpec', dict]):
        """
        Spec holds information about the request being evaluated
        """
        if isinstance(value, dict):
            value = TokenReviewSpec().from_dict(value)
        self._properties['spec'] = value

    @property
    def status(self) -> 'TokenReviewStatus':
        """
        Status is filled in by the server and indicates whether the
        request can be authenticated.
        """
        return self._properties.get('status')

    @status.setter
    def status(self, value: typing.Union['TokenReviewStatus', dict]):
        """
        Status is filled in by the server and indicates whether the
        request can be authenticated.
        """
        if isinstance(value, dict):
            value = TokenReviewStatus().from_dict(value)
        self._properties['status'] = value

    def create_resource(
            self,
            namespace: 'str' = None
    ) -> 'TokenReviewStatus':
        """
        Creates the TokenReview in the currently
        configured Kubernetes cluster and returns the status information
        returned by the Kubernetes API after the create is complete.
        """
        names = [
            'create_namespaced_token_review',
            'create_token_review'
        ]

        response = _kube_api.execute(
            action='create',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict()}
        )
        return (
            TokenReviewStatus()
            .from_dict(_kube_api.to_kuber_dict(response.status))
        )

    def replace_resource(
            self,
            namespace: 'str' = None
    ) -> 'TokenReviewStatus':
        """
        Replaces the TokenReview in the currently
        configured Kubernetes cluster and returns the status information
        returned by the Kubernetes API after the replace is complete.
        """
        names = [
            'replace_namespaced_token_review',
            'replace_token_review'
        ]

        response = _kube_api.execute(
            action='replace',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict(), 'name': self.metadata.name}
        )
        return (
            TokenReviewStatus()
            .from_dict(_kube_api.to_kuber_dict(response.status))
        )

    def patch_resource(
            self,
            namespace: 'str' = None
    ) -> 'TokenReviewStatus':
        """
        Patches the TokenReview in the currently
        configured Kubernetes cluster and returns the status information
        returned by the Kubernetes API after the replace is complete.
        """
        names = [
            'patch_namespaced_token_review',
            'patch_token_review'
        ]

        response = _kube_api.execute(
            action='patch',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict(), 'name': self.metadata.name}
        )
        return (
            TokenReviewStatus()
            .from_dict(_kube_api.to_kuber_dict(response.status))
        )

    def get_resource_status(
            self,
            namespace: 'str' = None
    ) -> 'TokenReviewStatus':
        """
        Returns status information about the given resource within the cluster.
        """
        names = [
            'read_namespaced_token_review',
            'read_token_review'
        ]

        response = _kube_api.execute(
            action='read',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'name': self.metadata.name}
        )
        return (
            TokenReviewStatus()
            .from_dict(_kube_api.to_kuber_dict(response.status))
        )

    def read_resource(
            self,
            namespace: str = None
    ):
        """
        Reads the TokenReview from the currently configured
        Kubernetes cluster and returns the low-level definition object.
        """
        names = [
            'read_namespaced_token_review',
            'read_token_review'
        ]
        return _kube_api.execute(
            action='read',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'name': self.metadata.name}
        )

    def delete_resource(
            self,
            namespace: str = None,
            propagation_policy: str = 'Foreground',
            grace_period_seconds: int = 10
    ):
        """
        Deletes the TokenReview from the currently configured
        Kubernetes cluster.
        """
        names = [
            'delete_namespaced_token_review',
            'delete_token_review'
        ]

        body = client.V1DeleteOptions(
            propagation_policy=propagation_policy,
            grace_period_seconds=grace_period_seconds
        )

        _kube_api.execute(
            action='delete',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'name': self.metadata.name, 'body': body}
        )

    @staticmethod
    def get_resource_api(
            api_client: client.ApiClient = None,
            **kwargs
    ) -> 'client.AuthenticationV1beta1Api':
        """
        Returns an instance of the kubernetes API client associated with
        this object.
        """
        if api_client:
            kwargs['apl_client'] = api_client
        return client.AuthenticationV1beta1Api(**kwargs)

    def __enter__(self) -> 'TokenReview':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class TokenReviewSpec(_kuber_definitions.Definition):
    """
    TokenReviewSpec is a description of the token authentication
    request.
    """

    def __init__(
            self,
            token: str = None,
    ):
        """Create TokenReviewSpec instance."""
        super(TokenReviewSpec, self).__init__(
            api_version='authentication/v1beta1',
            kind='TokenReviewSpec'
        )
        self._properties = {
            'token': token if token is not None else '',

        }
        self._types = {
            'token': (str, None),

        }

    @property
    def token(self) -> str:
        """
        Token is the opaque bearer token.
        """
        return self._properties.get('token')

    @token.setter
    def token(self, value: str):
        """
        Token is the opaque bearer token.
        """
        self._properties['token'] = value

    def __enter__(self) -> 'TokenReviewSpec':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class TokenReviewStatus(_kuber_definitions.Definition):
    """
    TokenReviewStatus is the result of the token authentication
    request.
    """

    def __init__(
            self,
            authenticated: bool = None,
            error: str = None,
            user: 'UserInfo' = None,
    ):
        """Create TokenReviewStatus instance."""
        super(TokenReviewStatus, self).__init__(
            api_version='authentication/v1beta1',
            kind='TokenReviewStatus'
        )
        self._properties = {
            'authenticated': authenticated if authenticated is not None else None,
            'error': error if error is not None else '',
            'user': user if user is not None else UserInfo(),

        }
        self._types = {
            'authenticated': (bool, None),
            'error': (str, None),
            'user': (UserInfo, None),

        }

    @property
    def authenticated(self) -> bool:
        """
        Authenticated indicates that the token was associated with a
        known user.
        """
        return self._properties.get('authenticated')

    @authenticated.setter
    def authenticated(self, value: bool):
        """
        Authenticated indicates that the token was associated with a
        known user.
        """
        self._properties['authenticated'] = value

    @property
    def error(self) -> str:
        """
        Error indicates that the token couldn't be checked
        """
        return self._properties.get('error')

    @error.setter
    def error(self, value: str):
        """
        Error indicates that the token couldn't be checked
        """
        self._properties['error'] = value

    @property
    def user(self) -> 'UserInfo':
        """
        User is the UserInfo associated with the provided token.
        """
        return self._properties.get('user')

    @user.setter
    def user(self, value: typing.Union['UserInfo', dict]):
        """
        User is the UserInfo associated with the provided token.
        """
        if isinstance(value, dict):
            value = UserInfo().from_dict(value)
        self._properties['user'] = value

    def __enter__(self) -> 'TokenReviewStatus':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class UserInfo(_kuber_definitions.Definition):
    """
    UserInfo holds the information about the user needed to
    implement the user.Info interface.
    """

    def __init__(
            self,
            extra: dict = None,
            groups: typing.List[str] = None,
            uid: str = None,
            username: str = None,
    ):
        """Create UserInfo instance."""
        super(UserInfo, self).__init__(
            api_version='authentication/v1beta1',
            kind='UserInfo'
        )
        self._properties = {
            'extra': extra if extra is not None else {},
            'groups': groups if groups is not None else [],
            'uid': uid if uid is not None else '',
            'username': username if username is not None else '',

        }
        self._types = {
            'extra': (dict, None),
            'groups': (list, str),
            'uid': (str, None),
            'username': (str, None),

        }

    @property
    def extra(self) -> dict:
        """
        Any additional information provided by the authenticator.
        """
        return self._properties.get('extra')

    @extra.setter
    def extra(self, value: dict):
        """
        Any additional information provided by the authenticator.
        """
        self._properties['extra'] = value

    @property
    def groups(self) -> typing.List[str]:
        """
        The names of groups this user is a part of.
        """
        return self._properties.get('groups')

    @groups.setter
    def groups(self, value: typing.List[str]):
        """
        The names of groups this user is a part of.
        """
        self._properties['groups'] = value

    @property
    def uid(self) -> str:
        """
        A unique value that identifies this user across time. If
        this user is deleted and another user by the same name is
        added, they will have different UIDs.
        """
        return self._properties.get('uid')

    @uid.setter
    def uid(self, value: str):
        """
        A unique value that identifies this user across time. If
        this user is deleted and another user by the same name is
        added, they will have different UIDs.
        """
        self._properties['uid'] = value

    @property
    def username(self) -> str:
        """
        The name that uniquely identifies this user among all active
        users.
        """
        return self._properties.get('username')

    @username.setter
    def username(self, value: str):
        """
        The name that uniquely identifies this user among all active
        users.
        """
        self._properties['username'] = value

    def __enter__(self) -> 'UserInfo':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False



import typing
import datetime as _datetime

from kubernetes import client
from kuber import kube_api as _kube_api

from kuber import definitions as _kuber_definitions
from kuber.v1_14.meta_v1 import ListMeta
from kuber.v1_14.meta_v1 import ObjectMeta
from kuber.v1_14.meta_v1 import Status
from kuber.v1_14.meta_v1 import StatusDetails


class CustomResourceColumnDefinition(_kuber_definitions.Definition):
    """
    CustomResourceColumnDefinition specifies a column for server
    side printing.
    """

    def __init__(
            self,
            jsonpath: str = None,
            description: str = None,
            format_: str = None,
            name: str = None,
            priority: int = None,
            type_: str = None,
    ):
        """Create CustomResourceColumnDefinition instance."""
        super(CustomResourceColumnDefinition, self).__init__(
            api_version='apiextensions/v1beta1',
            kind='CustomResourceColumnDefinition'
        )
        self._properties = {
            'JSONPath': jsonpath if jsonpath is not None else '',
            'description': description if description is not None else '',
            'format': format_ if format_ is not None else '',
            'name': name if name is not None else '',
            'priority': priority if priority is not None else None,
            'type': type_ if type_ is not None else '',

        }
        self._types = {
            'JSONPath': (str, None),
            'description': (str, None),
            'format': (str, None),
            'name': (str, None),
            'priority': (int, None),
            'type': (str, None),

        }

    @property
    def jsonpath(self) -> str:
        """
        JSONPath is a simple JSON path, i.e. with array notation.
        """
        return self._properties.get('JSONPath')

    @jsonpath.setter
    def jsonpath(self, value: str):
        """
        JSONPath is a simple JSON path, i.e. with array notation.
        """
        self._properties['JSONPath'] = value

    @property
    def description(self) -> str:
        """
        description is a human readable description of this column.
        """
        return self._properties.get('description')

    @description.setter
    def description(self, value: str):
        """
        description is a human readable description of this column.
        """
        self._properties['description'] = value

    @property
    def format_(self) -> str:
        """
        format is an optional OpenAPI type definition for this
        column. The 'name' format is applied to the primary
        identifier column to assist in clients identifying column is
        the resource name. See https://github.com/OAI/OpenAPI-
        Specification/blob/master/versions/2.0.md#data-types for
        more.
        """
        return self._properties.get('format')

    @format_.setter
    def format_(self, value: str):
        """
        format is an optional OpenAPI type definition for this
        column. The 'name' format is applied to the primary
        identifier column to assist in clients identifying column is
        the resource name. See https://github.com/OAI/OpenAPI-
        Specification/blob/master/versions/2.0.md#data-types for
        more.
        """
        self._properties['format'] = value

    @property
    def name(self) -> str:
        """
        name is a human readable name for the column.
        """
        return self._properties.get('name')

    @name.setter
    def name(self, value: str):
        """
        name is a human readable name for the column.
        """
        self._properties['name'] = value

    @property
    def priority(self) -> int:
        """
        priority is an integer defining the relative importance of
        this column compared to others. Lower numbers are considered
        higher priority. Columns that may be omitted in limited
        space scenarios should be given a higher priority.
        """
        return self._properties.get('priority')

    @priority.setter
    def priority(self, value: int):
        """
        priority is an integer defining the relative importance of
        this column compared to others. Lower numbers are considered
        higher priority. Columns that may be omitted in limited
        space scenarios should be given a higher priority.
        """
        self._properties['priority'] = value

    @property
    def type_(self) -> str:
        """
        type is an OpenAPI type definition for this column. See
        https://github.com/OAI/OpenAPI-
        Specification/blob/master/versions/2.0.md#data-types for
        more.
        """
        return self._properties.get('type')

    @type_.setter
    def type_(self, value: str):
        """
        type is an OpenAPI type definition for this column. See
        https://github.com/OAI/OpenAPI-
        Specification/blob/master/versions/2.0.md#data-types for
        more.
        """
        self._properties['type'] = value

    def __enter__(self) -> 'CustomResourceColumnDefinition':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class CustomResourceConversion(_kuber_definitions.Definition):
    """
    CustomResourceConversion describes how to convert different
    versions of a CR.
    """

    def __init__(
            self,
            conversion_review_versions: typing.List[str] = None,
            strategy: str = None,
            webhook_client_config: 'WebhookClientConfig' = None,
    ):
        """Create CustomResourceConversion instance."""
        super(CustomResourceConversion, self).__init__(
            api_version='apiextensions/v1beta1',
            kind='CustomResourceConversion'
        )
        self._properties = {
            'conversionReviewVersions': conversion_review_versions if conversion_review_versions is not None else [],
            'strategy': strategy if strategy is not None else '',
            'webhookClientConfig': webhook_client_config if webhook_client_config is not None else WebhookClientConfig(),

        }
        self._types = {
            'conversionReviewVersions': (list, str),
            'strategy': (str, None),
            'webhookClientConfig': (WebhookClientConfig, None),

        }

    @property
    def conversion_review_versions(self) -> typing.List[str]:
        """
        ConversionReviewVersions is an ordered list of preferred
        `ConversionReview` versions the Webhook expects. API server
        will try to use first version in the list which it supports.
        If none of the versions specified in this list supported by
        API server, conversion will fail for this object. If a
        persisted Webhook configuration specifies allowed versions
        and does not include any versions known to the API Server,
        calls to the webhook will fail. Default to `['v1beta1']`.
        """
        return self._properties.get('conversionReviewVersions')

    @conversion_review_versions.setter
    def conversion_review_versions(self, value: typing.List[str]):
        """
        ConversionReviewVersions is an ordered list of preferred
        `ConversionReview` versions the Webhook expects. API server
        will try to use first version in the list which it supports.
        If none of the versions specified in this list supported by
        API server, conversion will fail for this object. If a
        persisted Webhook configuration specifies allowed versions
        and does not include any versions known to the API Server,
        calls to the webhook will fail. Default to `['v1beta1']`.
        """
        self._properties['conversionReviewVersions'] = value

    @property
    def strategy(self) -> str:
        """
        `strategy` specifies the conversion strategy. Allowed values
        are: - `None`: The converter only change the apiVersion and
        would not touch any other field in the CR. - `Webhook`: API
        Server will call to an external webhook to do the
        conversion. Additional information is needed for this
        option.
        """
        return self._properties.get('strategy')

    @strategy.setter
    def strategy(self, value: str):
        """
        `strategy` specifies the conversion strategy. Allowed values
        are: - `None`: The converter only change the apiVersion and
        would not touch any other field in the CR. - `Webhook`: API
        Server will call to an external webhook to do the
        conversion. Additional information is needed for this
        option.
        """
        self._properties['strategy'] = value

    @property
    def webhook_client_config(self) -> 'WebhookClientConfig':
        """
        `webhookClientConfig` is the instructions for how to call
        the webhook if strategy is `Webhook`. This field is alpha-
        level and is only honored by servers that enable the
        CustomResourceWebhookConversion feature.
        """
        return self._properties.get('webhookClientConfig')

    @webhook_client_config.setter
    def webhook_client_config(self, value: typing.Union['WebhookClientConfig', dict]):
        """
        `webhookClientConfig` is the instructions for how to call
        the webhook if strategy is `Webhook`. This field is alpha-
        level and is only honored by servers that enable the
        CustomResourceWebhookConversion feature.
        """
        if isinstance(value, dict):
            value = WebhookClientConfig().from_dict(value)
        self._properties['webhookClientConfig'] = value

    def __enter__(self) -> 'CustomResourceConversion':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class CustomResourceDefinition(_kuber_definitions.Resource):
    """
    CustomResourceDefinition represents a resource that should
    be exposed on the API server.  Its name MUST be in the
    format <.spec.name>.<.spec.group>.
    """

    def __init__(
            self,
            metadata: 'ObjectMeta' = None,
            spec: 'CustomResourceDefinitionSpec' = None,
            status: 'CustomResourceDefinitionStatus' = None,
    ):
        """Create CustomResourceDefinition instance."""
        super(CustomResourceDefinition, self).__init__(
            api_version='apiextensions/v1beta1',
            kind='CustomResourceDefinition'
        )
        self._properties = {
            'metadata': metadata if metadata is not None else ObjectMeta(),
            'spec': spec if spec is not None else CustomResourceDefinitionSpec(),
            'status': status if status is not None else CustomResourceDefinitionStatus(),

        }
        self._types = {
            'apiVersion': (str, None),
            'kind': (str, None),
            'metadata': (ObjectMeta, None),
            'spec': (CustomResourceDefinitionSpec, None),
            'status': (CustomResourceDefinitionStatus, None),

        }

    @property
    def metadata(self) -> 'ObjectMeta':
        """

        """
        return self._properties.get('metadata')

    @metadata.setter
    def metadata(self, value: typing.Union['ObjectMeta', dict]):
        """

        """
        if isinstance(value, dict):
            value = ObjectMeta().from_dict(value)
        self._properties['metadata'] = value

    @property
    def spec(self) -> 'CustomResourceDefinitionSpec':
        """
        Spec describes how the user wants the resources to appear
        """
        return self._properties.get('spec')

    @spec.setter
    def spec(self, value: typing.Union['CustomResourceDefinitionSpec', dict]):
        """
        Spec describes how the user wants the resources to appear
        """
        if isinstance(value, dict):
            value = CustomResourceDefinitionSpec().from_dict(value)
        self._properties['spec'] = value

    @property
    def status(self) -> 'CustomResourceDefinitionStatus':
        """
        Status indicates the actual state of the
        CustomResourceDefinition
        """
        return self._properties.get('status')

    @status.setter
    def status(self, value: typing.Union['CustomResourceDefinitionStatus', dict]):
        """
        Status indicates the actual state of the
        CustomResourceDefinition
        """
        if isinstance(value, dict):
            value = CustomResourceDefinitionStatus().from_dict(value)
        self._properties['status'] = value

    def create_resource(
            self,
            namespace: 'str' = None
    ) -> 'CustomResourceDefinitionStatus':
        """
        Creates the CustomResourceDefinition in the currently
        configured Kubernetes cluster and returns the status information
        returned by the Kubernetes API after the create is complete.
        """
        names = [
            'create_namespaced_custom_resource_definition',
            'create_custom_resource_definition'
        ]

        response = _kube_api.execute(
            action='create',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict()}
        )
        return (
            CustomResourceDefinitionStatus()
            .from_dict(_kube_api.to_kuber_dict(response.status))
        )

    def replace_resource(
            self,
            namespace: 'str' = None
    ) -> 'CustomResourceDefinitionStatus':
        """
        Replaces the CustomResourceDefinition in the currently
        configured Kubernetes cluster and returns the status information
        returned by the Kubernetes API after the replace is complete.
        """
        names = [
            'replace_namespaced_custom_resource_definition',
            'replace_custom_resource_definition'
        ]

        response = _kube_api.execute(
            action='replace',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict(), 'name': self.metadata.name}
        )
        return (
            CustomResourceDefinitionStatus()
            .from_dict(_kube_api.to_kuber_dict(response.status))
        )

    def patch_resource(
            self,
            namespace: 'str' = None
    ) -> 'CustomResourceDefinitionStatus':
        """
        Patches the CustomResourceDefinition in the currently
        configured Kubernetes cluster and returns the status information
        returned by the Kubernetes API after the replace is complete.
        """
        names = [
            'patch_namespaced_custom_resource_definition',
            'patch_custom_resource_definition'
        ]

        response = _kube_api.execute(
            action='patch',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'body': self.to_dict(), 'name': self.metadata.name}
        )
        return (
            CustomResourceDefinitionStatus()
            .from_dict(_kube_api.to_kuber_dict(response.status))
        )

    def get_resource_status(
            self,
            namespace: 'str' = None
    ) -> 'CustomResourceDefinitionStatus':
        """
        Returns status information about the given resource within the cluster.
        """
        names = [
            'read_namespaced_custom_resource_definition',
            'read_custom_resource_definition'
        ]

        response = _kube_api.execute(
            action='read',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'name': self.metadata.name}
        )
        return (
            CustomResourceDefinitionStatus()
            .from_dict(_kube_api.to_kuber_dict(response.status))
        )

    def read_resource(
            self,
            namespace: str = None
    ):
        """
        Reads the CustomResourceDefinition from the currently configured
        Kubernetes cluster and returns the low-level definition object.
        """
        names = [
            'read_namespaced_custom_resource_definition',
            'read_custom_resource_definition'
        ]
        return _kube_api.execute(
            action='read',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'name': self.metadata.name}
        )

    def delete_resource(
            self,
            namespace: str = None,
            propagation_policy: str = 'Foreground',
            grace_period_seconds: int = 10
    ):
        """
        Deletes the CustomResourceDefinition from the currently configured
        Kubernetes cluster.
        """
        names = [
            'delete_namespaced_custom_resource_definition',
            'delete_custom_resource_definition'
        ]

        body = client.V1DeleteOptions(
            propagation_policy=propagation_policy,
            grace_period_seconds=grace_period_seconds
        )

        _kube_api.execute(
            action='delete',
            resource=self,
            names=names,
            namespace=namespace,
            api_client=None,
            api_args={'name': self.metadata.name, 'body': body}
        )

    @staticmethod
    def get_resource_api(
            api_client: client.ApiClient = None,
            **kwargs
    ) -> 'client.ApiextensionsV1beta1Api':
        """
        Returns an instance of the kubernetes API client associated with
        this object.
        """
        if api_client:
            kwargs['apl_client'] = api_client
        return client.ApiextensionsV1beta1Api(**kwargs)

    def __enter__(self) -> 'CustomResourceDefinition':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class CustomResourceDefinitionCondition(_kuber_definitions.Definition):
    """
    CustomResourceDefinitionCondition contains details for the
    current condition of this pod.
    """

    def __init__(
            self,
            last_transition_time: str = None,
            message: str = None,
            reason: str = None,
            status: str = None,
            type_: str = None,
    ):
        """Create CustomResourceDefinitionCondition instance."""
        super(CustomResourceDefinitionCondition, self).__init__(
            api_version='apiextensions/v1beta1',
            kind='CustomResourceDefinitionCondition'
        )
        self._properties = {
            'lastTransitionTime': last_transition_time if last_transition_time is not None else None,
            'message': message if message is not None else '',
            'reason': reason if reason is not None else '',
            'status': status if status is not None else '',
            'type': type_ if type_ is not None else '',

        }
        self._types = {
            'lastTransitionTime': (str, None),
            'message': (str, None),
            'reason': (str, None),
            'status': (str, None),
            'type': (str, None),

        }

    @property
    def last_transition_time(self) -> str:
        """
        Last time the condition transitioned from one status to
        another.
        """
        return self._properties.get('lastTransitionTime')

    @last_transition_time.setter
    def last_transition_time(
            self,
            value: typing.Union[str, _datetime.datetime, _datetime.date]
    ):
        """
        Last time the condition transitioned from one status to
        another.
        """
        if isinstance(value, _datetime.datetime):
            value = value.strftime('%Y-%m-%dT%H:%M:%SZ')
        elif isinstance(value, _datetime.date):
            value = value.strftime('%Y-%m-%dT00:00:00Z')
        self._properties['lastTransitionTime'] = value

    @property
    def message(self) -> str:
        """
        Human-readable message indicating details about last
        transition.
        """
        return self._properties.get('message')

    @message.setter
    def message(self, value: str):
        """
        Human-readable message indicating details about last
        transition.
        """
        self._properties['message'] = value

    @property
    def reason(self) -> str:
        """
        Unique, one-word, CamelCase reason for the condition's last
        transition.
        """
        return self._properties.get('reason')

    @reason.setter
    def reason(self, value: str):
        """
        Unique, one-word, CamelCase reason for the condition's last
        transition.
        """
        self._properties['reason'] = value

    @property
    def status(self) -> str:
        """
        Status is the status of the condition. Can be True, False,
        Unknown.
        """
        return self._properties.get('status')

    @status.setter
    def status(self, value: str):
        """
        Status is the status of the condition. Can be True, False,
        Unknown.
        """
        self._properties['status'] = value

    @property
    def type_(self) -> str:
        """
        Type is the type of the condition.
        """
        return self._properties.get('type')

    @type_.setter
    def type_(self, value: str):
        """
        Type is the type of the condition.
        """
        self._properties['type'] = value

    def __enter__(self) -> 'CustomResourceDefinitionCondition':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class CustomResourceDefinitionList(_kuber_definitions.Collection):
    """
    CustomResourceDefinitionList is a list of
    CustomResourceDefinition objects.
    """

    def __init__(
            self,
            items: typing.List['CustomResourceDefinition'] = None,
            metadata: 'ListMeta' = None,
    ):
        """Create CustomResourceDefinitionList instance."""
        super(CustomResourceDefinitionList, self).__init__(
            api_version='apiextensions/v1beta1',
            kind='CustomResourceDefinitionList'
        )
        self._properties = {
            'items': items if items is not None else [],
            'metadata': metadata if metadata is not None else ListMeta(),

        }
        self._types = {
            'apiVersion': (str, None),
            'items': (list, CustomResourceDefinition),
            'kind': (str, None),
            'metadata': (ListMeta, None),

        }

    @property
    def items(self) -> typing.List['CustomResourceDefinition']:
        """
        Items individual CustomResourceDefinitions
        """
        return self._properties.get('items')

    @items.setter
    def items(
            self,
            value: typing.Union[typing.List['CustomResourceDefinition'], typing.List[dict]]
    ):
        """
        Items individual CustomResourceDefinitions
        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = CustomResourceDefinition().from_dict(item)
            cleaned.append(item)
        self._properties['items'] = cleaned

    @property
    def metadata(self) -> 'ListMeta':
        """

        """
        return self._properties.get('metadata')

    @metadata.setter
    def metadata(self, value: typing.Union['ListMeta', dict]):
        """

        """
        if isinstance(value, dict):
            value = ListMeta().from_dict(value)
        self._properties['metadata'] = value

    @staticmethod
    def get_resource_api(
            api_client: client.ApiClient = None,
            **kwargs
    ) -> 'client.ApiextensionsV1beta1Api':
        """
        Returns an instance of the kubernetes API client associated with
        this object.
        """
        if api_client:
            kwargs['apl_client'] = api_client
        return client.ApiextensionsV1beta1Api(**kwargs)

    def __enter__(self) -> 'CustomResourceDefinitionList':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class CustomResourceDefinitionNames(_kuber_definitions.Definition):
    """
    CustomResourceDefinitionNames indicates the names to serve
    this CustomResourceDefinition
    """

    def __init__(
            self,
            categories: typing.List[str] = None,
            kind: str = None,
            list_kind: str = None,
            plural: str = None,
            short_names: typing.List[str] = None,
            singular: str = None,
    ):
        """Create CustomResourceDefinitionNames instance."""
        super(CustomResourceDefinitionNames, self).__init__(
            api_version='apiextensions/v1beta1',
            kind='CustomResourceDefinitionNames'
        )
        self._properties = {
            'categories': categories if categories is not None else [],
            'kind': kind if kind is not None else '',
            'listKind': list_kind if list_kind is not None else '',
            'plural': plural if plural is not None else '',
            'shortNames': short_names if short_names is not None else [],
            'singular': singular if singular is not None else '',

        }
        self._types = {
            'categories': (list, str),
            'kind': (str, None),
            'listKind': (str, None),
            'plural': (str, None),
            'shortNames': (list, str),
            'singular': (str, None),

        }

    @property
    def categories(self) -> typing.List[str]:
        """
        Categories is a list of grouped resources custom resources
        belong to (e.g. 'all')
        """
        return self._properties.get('categories')

    @categories.setter
    def categories(self, value: typing.List[str]):
        """
        Categories is a list of grouped resources custom resources
        belong to (e.g. 'all')
        """
        self._properties['categories'] = value

    @property
    def kind(self) -> str:
        """
        Kind is the serialized kind of the resource.  It is normally
        CamelCase and singular.
        """
        return self._properties.get('kind')

    @kind.setter
    def kind(self, value: str):
        """
        Kind is the serialized kind of the resource.  It is normally
        CamelCase and singular.
        """
        self._properties['kind'] = value

    @property
    def list_kind(self) -> str:
        """
        ListKind is the serialized kind of the list for this
        resource.  Defaults to <kind>List.
        """
        return self._properties.get('listKind')

    @list_kind.setter
    def list_kind(self, value: str):
        """
        ListKind is the serialized kind of the list for this
        resource.  Defaults to <kind>List.
        """
        self._properties['listKind'] = value

    @property
    def plural(self) -> str:
        """
        Plural is the plural name of the resource to serve.  It must
        match the name of the CustomResourceDefinition-registration
        too: plural.group and it must be all lowercase.
        """
        return self._properties.get('plural')

    @plural.setter
    def plural(self, value: str):
        """
        Plural is the plural name of the resource to serve.  It must
        match the name of the CustomResourceDefinition-registration
        too: plural.group and it must be all lowercase.
        """
        self._properties['plural'] = value

    @property
    def short_names(self) -> typing.List[str]:
        """
        ShortNames are short names for the resource.  It must be all
        lowercase.
        """
        return self._properties.get('shortNames')

    @short_names.setter
    def short_names(self, value: typing.List[str]):
        """
        ShortNames are short names for the resource.  It must be all
        lowercase.
        """
        self._properties['shortNames'] = value

    @property
    def singular(self) -> str:
        """
        Singular is the singular name of the resource.  It must be
        all lowercase  Defaults to lowercased <kind>
        """
        return self._properties.get('singular')

    @singular.setter
    def singular(self, value: str):
        """
        Singular is the singular name of the resource.  It must be
        all lowercase  Defaults to lowercased <kind>
        """
        self._properties['singular'] = value

    def __enter__(self) -> 'CustomResourceDefinitionNames':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class CustomResourceDefinitionSpec(_kuber_definitions.Definition):
    """
    CustomResourceDefinitionSpec describes how a user wants
    their resource to appear
    """

    def __init__(
            self,
            additional_printer_columns: typing.List['CustomResourceColumnDefinition'] = None,
            conversion: 'CustomResourceConversion' = None,
            group: str = None,
            names: 'CustomResourceDefinitionNames' = None,
            scope: str = None,
            subresources: 'CustomResourceSubresources' = None,
            validation: 'CustomResourceValidation' = None,
            version: str = None,
            versions: typing.List['CustomResourceDefinitionVersion'] = None,
    ):
        """Create CustomResourceDefinitionSpec instance."""
        super(CustomResourceDefinitionSpec, self).__init__(
            api_version='apiextensions/v1beta1',
            kind='CustomResourceDefinitionSpec'
        )
        self._properties = {
            'additionalPrinterColumns': additional_printer_columns if additional_printer_columns is not None else [],
            'conversion': conversion if conversion is not None else CustomResourceConversion(),
            'group': group if group is not None else '',
            'names': names if names is not None else CustomResourceDefinitionNames(),
            'scope': scope if scope is not None else '',
            'subresources': subresources if subresources is not None else CustomResourceSubresources(),
            'validation': validation if validation is not None else CustomResourceValidation(),
            'version': version if version is not None else '',
            'versions': versions if versions is not None else [],

        }
        self._types = {
            'additionalPrinterColumns': (list, CustomResourceColumnDefinition),
            'conversion': (CustomResourceConversion, None),
            'group': (str, None),
            'names': (CustomResourceDefinitionNames, None),
            'scope': (str, None),
            'subresources': (CustomResourceSubresources, None),
            'validation': (CustomResourceValidation, None),
            'version': (str, None),
            'versions': (list, CustomResourceDefinitionVersion),

        }

    @property
    def additional_printer_columns(self) -> typing.List['CustomResourceColumnDefinition']:
        """
        AdditionalPrinterColumns are additional columns shown e.g.
        in kubectl next to the name. Defaults to a created-at
        column. Optional, the global columns for all versions. Top-
        level and per-version columns are mutually exclusive.
        """
        return self._properties.get('additionalPrinterColumns')

    @additional_printer_columns.setter
    def additional_printer_columns(
            self,
            value: typing.Union[typing.List['CustomResourceColumnDefinition'], typing.List[dict]]
    ):
        """
        AdditionalPrinterColumns are additional columns shown e.g.
        in kubectl next to the name. Defaults to a created-at
        column. Optional, the global columns for all versions. Top-
        level and per-version columns are mutually exclusive.
        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = CustomResourceColumnDefinition().from_dict(item)
            cleaned.append(item)
        self._properties['additionalPrinterColumns'] = cleaned

    @property
    def conversion(self) -> 'CustomResourceConversion':
        """
        `conversion` defines conversion settings for the CRD.
        """
        return self._properties.get('conversion')

    @conversion.setter
    def conversion(self, value: typing.Union['CustomResourceConversion', dict]):
        """
        `conversion` defines conversion settings for the CRD.
        """
        if isinstance(value, dict):
            value = CustomResourceConversion().from_dict(value)
        self._properties['conversion'] = value

    @property
    def group(self) -> str:
        """
        Group is the group this resource belongs in
        """
        return self._properties.get('group')

    @group.setter
    def group(self, value: str):
        """
        Group is the group this resource belongs in
        """
        self._properties['group'] = value

    @property
    def names(self) -> 'CustomResourceDefinitionNames':
        """
        Names are the names used to describe this custom resource
        """
        return self._properties.get('names')

    @names.setter
    def names(self, value: typing.Union['CustomResourceDefinitionNames', dict]):
        """
        Names are the names used to describe this custom resource
        """
        if isinstance(value, dict):
            value = CustomResourceDefinitionNames().from_dict(value)
        self._properties['names'] = value

    @property
    def scope(self) -> str:
        """
        Scope indicates whether this resource is cluster or
        namespace scoped.  Default is namespaced
        """
        return self._properties.get('scope')

    @scope.setter
    def scope(self, value: str):
        """
        Scope indicates whether this resource is cluster or
        namespace scoped.  Default is namespaced
        """
        self._properties['scope'] = value

    @property
    def subresources(self) -> 'CustomResourceSubresources':
        """
        Subresources describes the subresources for CustomResource
        Optional, the global subresources for all versions. Top-
        level and per-version subresources are mutually exclusive.
        """
        return self._properties.get('subresources')

    @subresources.setter
    def subresources(self, value: typing.Union['CustomResourceSubresources', dict]):
        """
        Subresources describes the subresources for CustomResource
        Optional, the global subresources for all versions. Top-
        level and per-version subresources are mutually exclusive.
        """
        if isinstance(value, dict):
            value = CustomResourceSubresources().from_dict(value)
        self._properties['subresources'] = value

    @property
    def validation(self) -> 'CustomResourceValidation':
        """
        Validation describes the validation methods for
        CustomResources Optional, the global validation schema for
        all versions. Top-level and per-version schemas are mutually
        exclusive.
        """
        return self._properties.get('validation')

    @validation.setter
    def validation(self, value: typing.Union['CustomResourceValidation', dict]):
        """
        Validation describes the validation methods for
        CustomResources Optional, the global validation schema for
        all versions. Top-level and per-version schemas are mutually
        exclusive.
        """
        if isinstance(value, dict):
            value = CustomResourceValidation().from_dict(value)
        self._properties['validation'] = value

    @property
    def version(self) -> str:
        """
        Version is the version this resource belongs in Should be
        always first item in Versions field if provided. Optional,
        but at least one of Version or Versions must be set.
        Deprecated: Please use `Versions`.
        """
        return self._properties.get('version')

    @version.setter
    def version(self, value: str):
        """
        Version is the version this resource belongs in Should be
        always first item in Versions field if provided. Optional,
        but at least one of Version or Versions must be set.
        Deprecated: Please use `Versions`.
        """
        self._properties['version'] = value

    @property
    def versions(self) -> typing.List['CustomResourceDefinitionVersion']:
        """
        Versions is the list of all supported versions for this
        resource. If Version field is provided, this field is
        optional. Validation: All versions must use the same
        validation schema for now. i.e., top level Validation field
        is applied to all of these versions. Order: The version name
        will be used to compute the order. If the version string is
        "kube-like", it will sort above non "kube-like" version
        strings, which are ordered lexicographically. "Kube-like"
        versions start with a "v", then are followed by a number
        (the major version), then optionally the string "alpha" or
        "beta" and another number (the minor version). These are
        sorted first by GA > beta > alpha (where GA is a version
        with no suffix such as beta or alpha), and then by comparing
        major version, then minor version. An example sorted list of
        versions: v10, v2, v1, v11beta2, v10beta3, v3beta1,
        v12alpha1, v11alpha2, foo1, foo10.
        """
        return self._properties.get('versions')

    @versions.setter
    def versions(
            self,
            value: typing.Union[typing.List['CustomResourceDefinitionVersion'], typing.List[dict]]
    ):
        """
        Versions is the list of all supported versions for this
        resource. If Version field is provided, this field is
        optional. Validation: All versions must use the same
        validation schema for now. i.e., top level Validation field
        is applied to all of these versions. Order: The version name
        will be used to compute the order. If the version string is
        "kube-like", it will sort above non "kube-like" version
        strings, which are ordered lexicographically. "Kube-like"
        versions start with a "v", then are followed by a number
        (the major version), then optionally the string "alpha" or
        "beta" and another number (the minor version). These are
        sorted first by GA > beta > alpha (where GA is a version
        with no suffix such as beta or alpha), and then by comparing
        major version, then minor version. An example sorted list of
        versions: v10, v2, v1, v11beta2, v10beta3, v3beta1,
        v12alpha1, v11alpha2, foo1, foo10.
        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = CustomResourceDefinitionVersion().from_dict(item)
            cleaned.append(item)
        self._properties['versions'] = cleaned

    def __enter__(self) -> 'CustomResourceDefinitionSpec':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class CustomResourceDefinitionStatus(_kuber_definitions.Definition):
    """
    CustomResourceDefinitionStatus indicates the state of the
    CustomResourceDefinition
    """

    def __init__(
            self,
            accepted_names: 'CustomResourceDefinitionNames' = None,
            conditions: typing.List['CustomResourceDefinitionCondition'] = None,
            stored_versions: typing.List[str] = None,
    ):
        """Create CustomResourceDefinitionStatus instance."""
        super(CustomResourceDefinitionStatus, self).__init__(
            api_version='apiextensions/v1beta1',
            kind='CustomResourceDefinitionStatus'
        )
        self._properties = {
            'acceptedNames': accepted_names if accepted_names is not None else CustomResourceDefinitionNames(),
            'conditions': conditions if conditions is not None else [],
            'storedVersions': stored_versions if stored_versions is not None else [],

        }
        self._types = {
            'acceptedNames': (CustomResourceDefinitionNames, None),
            'conditions': (list, CustomResourceDefinitionCondition),
            'storedVersions': (list, str),

        }

    @property
    def accepted_names(self) -> 'CustomResourceDefinitionNames':
        """
        AcceptedNames are the names that are actually being used to
        serve discovery They may be different than the names in
        spec.
        """
        return self._properties.get('acceptedNames')

    @accepted_names.setter
    def accepted_names(self, value: typing.Union['CustomResourceDefinitionNames', dict]):
        """
        AcceptedNames are the names that are actually being used to
        serve discovery They may be different than the names in
        spec.
        """
        if isinstance(value, dict):
            value = CustomResourceDefinitionNames().from_dict(value)
        self._properties['acceptedNames'] = value

    @property
    def conditions(self) -> typing.List['CustomResourceDefinitionCondition']:
        """
        Conditions indicate state for particular aspects of a
        CustomResourceDefinition
        """
        return self._properties.get('conditions')

    @conditions.setter
    def conditions(
            self,
            value: typing.Union[typing.List['CustomResourceDefinitionCondition'], typing.List[dict]]
    ):
        """
        Conditions indicate state for particular aspects of a
        CustomResourceDefinition
        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = CustomResourceDefinitionCondition().from_dict(item)
            cleaned.append(item)
        self._properties['conditions'] = cleaned

    @property
    def stored_versions(self) -> typing.List[str]:
        """
        StoredVersions are all versions of CustomResources that were
        ever persisted. Tracking these versions allows a migration
        path for stored versions in etcd. The field is mutable so
        the migration controller can first finish a migration to
        another version (i.e. that no old objects are left in the
        storage), and then remove the rest of the versions from this
        list. None of the versions in this list can be removed from
        the spec.Versions field.
        """
        return self._properties.get('storedVersions')

    @stored_versions.setter
    def stored_versions(self, value: typing.List[str]):
        """
        StoredVersions are all versions of CustomResources that were
        ever persisted. Tracking these versions allows a migration
        path for stored versions in etcd. The field is mutable so
        the migration controller can first finish a migration to
        another version (i.e. that no old objects are left in the
        storage), and then remove the rest of the versions from this
        list. None of the versions in this list can be removed from
        the spec.Versions field.
        """
        self._properties['storedVersions'] = value

    def __enter__(self) -> 'CustomResourceDefinitionStatus':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class CustomResourceDefinitionVersion(_kuber_definitions.Definition):
    """
    CustomResourceDefinitionVersion describes a version for CRD.
    """

    def __init__(
            self,
            additional_printer_columns: typing.List['CustomResourceColumnDefinition'] = None,
            name: str = None,
            schema: 'CustomResourceValidation' = None,
            served: bool = None,
            storage: bool = None,
            subresources: 'CustomResourceSubresources' = None,
    ):
        """Create CustomResourceDefinitionVersion instance."""
        super(CustomResourceDefinitionVersion, self).__init__(
            api_version='apiextensions/v1beta1',
            kind='CustomResourceDefinitionVersion'
        )
        self._properties = {
            'additionalPrinterColumns': additional_printer_columns if additional_printer_columns is not None else [],
            'name': name if name is not None else '',
            'schema': schema if schema is not None else CustomResourceValidation(),
            'served': served if served is not None else None,
            'storage': storage if storage is not None else None,
            'subresources': subresources if subresources is not None else CustomResourceSubresources(),

        }
        self._types = {
            'additionalPrinterColumns': (list, CustomResourceColumnDefinition),
            'name': (str, None),
            'schema': (CustomResourceValidation, None),
            'served': (bool, None),
            'storage': (bool, None),
            'subresources': (CustomResourceSubresources, None),

        }

    @property
    def additional_printer_columns(self) -> typing.List['CustomResourceColumnDefinition']:
        """
        AdditionalPrinterColumns are additional columns shown e.g.
        in kubectl next to the name. Defaults to a created-at
        column. Top-level and per-version columns are mutually
        exclusive. Per-version columns must not all be set to
        identical values (top-level columns should be used instead)
        This field is alpha-level and is only honored by servers
        that enable the CustomResourceWebhookConversion feature.
        NOTE: CRDs created prior to 1.13 populated the top-level
        additionalPrinterColumns field by default. To apply an
        update that changes to per-version additionalPrinterColumns,
        the top-level additionalPrinterColumns field must be
        explicitly set to null
        """
        return self._properties.get('additionalPrinterColumns')

    @additional_printer_columns.setter
    def additional_printer_columns(
            self,
            value: typing.Union[typing.List['CustomResourceColumnDefinition'], typing.List[dict]]
    ):
        """
        AdditionalPrinterColumns are additional columns shown e.g.
        in kubectl next to the name. Defaults to a created-at
        column. Top-level and per-version columns are mutually
        exclusive. Per-version columns must not all be set to
        identical values (top-level columns should be used instead)
        This field is alpha-level and is only honored by servers
        that enable the CustomResourceWebhookConversion feature.
        NOTE: CRDs created prior to 1.13 populated the top-level
        additionalPrinterColumns field by default. To apply an
        update that changes to per-version additionalPrinterColumns,
        the top-level additionalPrinterColumns field must be
        explicitly set to null
        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = CustomResourceColumnDefinition().from_dict(item)
            cleaned.append(item)
        self._properties['additionalPrinterColumns'] = cleaned

    @property
    def name(self) -> str:
        """
        Name is the version name, e.g. v1, v2beta1, etc.
        """
        return self._properties.get('name')

    @name.setter
    def name(self, value: str):
        """
        Name is the version name, e.g. v1, v2beta1, etc.
        """
        self._properties['name'] = value

    @property
    def schema(self) -> 'CustomResourceValidation':
        """
        Schema describes the schema for CustomResource used in
        validation, pruning, and defaulting. Top-level and per-
        version schemas are mutually exclusive. Per-version schemas
        must not all be set to identical values (top-level
        validation schema should be used instead) This field is
        alpha-level and is only honored by servers that enable the
        CustomResourceWebhookConversion feature.
        """
        return self._properties.get('schema')

    @schema.setter
    def schema(self, value: typing.Union['CustomResourceValidation', dict]):
        """
        Schema describes the schema for CustomResource used in
        validation, pruning, and defaulting. Top-level and per-
        version schemas are mutually exclusive. Per-version schemas
        must not all be set to identical values (top-level
        validation schema should be used instead) This field is
        alpha-level and is only honored by servers that enable the
        CustomResourceWebhookConversion feature.
        """
        if isinstance(value, dict):
            value = CustomResourceValidation().from_dict(value)
        self._properties['schema'] = value

    @property
    def served(self) -> bool:
        """
        Served is a flag enabling/disabling this version from being
        served via REST APIs
        """
        return self._properties.get('served')

    @served.setter
    def served(self, value: bool):
        """
        Served is a flag enabling/disabling this version from being
        served via REST APIs
        """
        self._properties['served'] = value

    @property
    def storage(self) -> bool:
        """
        Storage flags the version as storage version. There must be
        exactly one flagged as storage version.
        """
        return self._properties.get('storage')

    @storage.setter
    def storage(self, value: bool):
        """
        Storage flags the version as storage version. There must be
        exactly one flagged as storage version.
        """
        self._properties['storage'] = value

    @property
    def subresources(self) -> 'CustomResourceSubresources':
        """
        Subresources describes the subresources for CustomResource
        Top-level and per-version subresources are mutually
        exclusive. Per-version subresources must not all be set to
        identical values (top-level subresources should be used
        instead) This field is alpha-level and is only honored by
        servers that enable the CustomResourceWebhookConversion
        feature.
        """
        return self._properties.get('subresources')

    @subresources.setter
    def subresources(self, value: typing.Union['CustomResourceSubresources', dict]):
        """
        Subresources describes the subresources for CustomResource
        Top-level and per-version subresources are mutually
        exclusive. Per-version subresources must not all be set to
        identical values (top-level subresources should be used
        instead) This field is alpha-level and is only honored by
        servers that enable the CustomResourceWebhookConversion
        feature.
        """
        if isinstance(value, dict):
            value = CustomResourceSubresources().from_dict(value)
        self._properties['subresources'] = value

    def __enter__(self) -> 'CustomResourceDefinitionVersion':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class CustomResourceSubresourceScale(_kuber_definitions.Definition):
    """
    CustomResourceSubresourceScale defines how to serve the
    scale subresource for CustomResources.
    """

    def __init__(
            self,
            label_selector_path: str = None,
            spec_replicas_path: str = None,
            status_replicas_path: str = None,
    ):
        """Create CustomResourceSubresourceScale instance."""
        super(CustomResourceSubresourceScale, self).__init__(
            api_version='apiextensions/v1beta1',
            kind='CustomResourceSubresourceScale'
        )
        self._properties = {
            'labelSelectorPath': label_selector_path if label_selector_path is not None else '',
            'specReplicasPath': spec_replicas_path if spec_replicas_path is not None else '',
            'statusReplicasPath': status_replicas_path if status_replicas_path is not None else '',

        }
        self._types = {
            'labelSelectorPath': (str, None),
            'specReplicasPath': (str, None),
            'statusReplicasPath': (str, None),

        }

    @property
    def label_selector_path(self) -> str:
        """
        LabelSelectorPath defines the JSON path inside of a
        CustomResource that corresponds to Scale.Status.Selector.
        Only JSON paths without the array notation are allowed. Must
        be a JSON Path under .status. Must be set to work with HPA.
        If there is no value under the given path in the
        CustomResource, the status label selector value in the
        /scale subresource will default to the empty string.
        """
        return self._properties.get('labelSelectorPath')

    @label_selector_path.setter
    def label_selector_path(self, value: str):
        """
        LabelSelectorPath defines the JSON path inside of a
        CustomResource that corresponds to Scale.Status.Selector.
        Only JSON paths without the array notation are allowed. Must
        be a JSON Path under .status. Must be set to work with HPA.
        If there is no value under the given path in the
        CustomResource, the status label selector value in the
        /scale subresource will default to the empty string.
        """
        self._properties['labelSelectorPath'] = value

    @property
    def spec_replicas_path(self) -> str:
        """
        SpecReplicasPath defines the JSON path inside of a
        CustomResource that corresponds to Scale.Spec.Replicas. Only
        JSON paths without the array notation are allowed. Must be a
        JSON Path under .spec. If there is no value under the given
        path in the CustomResource, the /scale subresource will
        return an error on GET.
        """
        return self._properties.get('specReplicasPath')

    @spec_replicas_path.setter
    def spec_replicas_path(self, value: str):
        """
        SpecReplicasPath defines the JSON path inside of a
        CustomResource that corresponds to Scale.Spec.Replicas. Only
        JSON paths without the array notation are allowed. Must be a
        JSON Path under .spec. If there is no value under the given
        path in the CustomResource, the /scale subresource will
        return an error on GET.
        """
        self._properties['specReplicasPath'] = value

    @property
    def status_replicas_path(self) -> str:
        """
        StatusReplicasPath defines the JSON path inside of a
        CustomResource that corresponds to Scale.Status.Replicas.
        Only JSON paths without the array notation are allowed. Must
        be a JSON Path under .status. If there is no value under the
        given path in the CustomResource, the status replica value
        in the /scale subresource will default to 0.
        """
        return self._properties.get('statusReplicasPath')

    @status_replicas_path.setter
    def status_replicas_path(self, value: str):
        """
        StatusReplicasPath defines the JSON path inside of a
        CustomResource that corresponds to Scale.Status.Replicas.
        Only JSON paths without the array notation are allowed. Must
        be a JSON Path under .status. If there is no value under the
        given path in the CustomResource, the status replica value
        in the /scale subresource will default to 0.
        """
        self._properties['statusReplicasPath'] = value

    def __enter__(self) -> 'CustomResourceSubresourceScale':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class CustomResourceSubresourceStatus(_kuber_definitions.Definition):
    """
    CustomResourceSubresourceStatus defines how to serve the
    status subresource for CustomResources. Status is
    represented by the `.status` JSON path inside of a
    CustomResource. When set, * exposes a /status subresource
    for the custom resource * PUT requests to the /status
    subresource take a custom resource object, and ignore
    changes to anything except the status stanza *
    PUT/POST/PATCH requests to the custom resource ignore
    changes to the status stanza
    """

    def __init__(
            self,
    ):
        """Create CustomResourceSubresourceStatus instance."""
        super(CustomResourceSubresourceStatus, self).__init__(
            api_version='apiextensions/v1beta1',
            kind='CustomResourceSubresourceStatus'
        )
        self._properties = {

        }
        self._types = {

        }

    def __enter__(self) -> 'CustomResourceSubresourceStatus':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class CustomResourceSubresources(_kuber_definitions.Definition):
    """
    CustomResourceSubresources defines the status and scale
    subresources for CustomResources.
    """

    def __init__(
            self,
            scale: 'CustomResourceSubresourceScale' = None,
            status: 'CustomResourceSubresourceStatus' = None,
    ):
        """Create CustomResourceSubresources instance."""
        super(CustomResourceSubresources, self).__init__(
            api_version='apiextensions/v1beta1',
            kind='CustomResourceSubresources'
        )
        self._properties = {
            'scale': scale if scale is not None else CustomResourceSubresourceScale(),
            'status': status if status is not None else CustomResourceSubresourceStatus(),

        }
        self._types = {
            'scale': (CustomResourceSubresourceScale, None),
            'status': (CustomResourceSubresourceStatus, None),

        }

    @property
    def scale(self) -> 'CustomResourceSubresourceScale':
        """
        Scale denotes the scale subresource for CustomResources
        """
        return self._properties.get('scale')

    @scale.setter
    def scale(self, value: typing.Union['CustomResourceSubresourceScale', dict]):
        """
        Scale denotes the scale subresource for CustomResources
        """
        if isinstance(value, dict):
            value = CustomResourceSubresourceScale().from_dict(value)
        self._properties['scale'] = value

    @property
    def status(self) -> 'CustomResourceSubresourceStatus':
        """
        Status denotes the status subresource for CustomResources
        """
        return self._properties.get('status')

    @status.setter
    def status(self, value: typing.Union['CustomResourceSubresourceStatus', dict]):
        """
        Status denotes the status subresource for CustomResources
        """
        if isinstance(value, dict):
            value = CustomResourceSubresourceStatus().from_dict(value)
        self._properties['status'] = value

    def __enter__(self) -> 'CustomResourceSubresources':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class CustomResourceValidation(_kuber_definitions.Definition):
    """
    CustomResourceValidation is a list of validation methods for
    CustomResources.
    """

    def __init__(
            self,
            open_apiv3_schema: 'JSONSchemaProps' = None,
    ):
        """Create CustomResourceValidation instance."""
        super(CustomResourceValidation, self).__init__(
            api_version='apiextensions/v1beta1',
            kind='CustomResourceValidation'
        )
        self._properties = {
            'openAPIV3Schema': open_apiv3_schema if open_apiv3_schema is not None else JSONSchemaProps(),

        }
        self._types = {
            'openAPIV3Schema': (JSONSchemaProps, None),

        }

    @property
    def open_apiv3_schema(self) -> 'JSONSchemaProps':
        """
        OpenAPIV3Schema is the OpenAPI v3 schema to be validated
        against.
        """
        return self._properties.get('openAPIV3Schema')

    @open_apiv3_schema.setter
    def open_apiv3_schema(self, value: typing.Union['JSONSchemaProps', dict]):
        """
        OpenAPIV3Schema is the OpenAPI v3 schema to be validated
        against.
        """
        if isinstance(value, dict):
            value = JSONSchemaProps().from_dict(value)
        self._properties['openAPIV3Schema'] = value

    def __enter__(self) -> 'CustomResourceValidation':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class ExternalDocumentation(_kuber_definitions.Definition):
    """
    ExternalDocumentation allows referencing an external
    resource for extended documentation.
    """

    def __init__(
            self,
            description: str = None,
            url: str = None,
    ):
        """Create ExternalDocumentation instance."""
        super(ExternalDocumentation, self).__init__(
            api_version='apiextensions/v1beta1',
            kind='ExternalDocumentation'
        )
        self._properties = {
            'description': description if description is not None else '',
            'url': url if url is not None else '',

        }
        self._types = {
            'description': (str, None),
            'url': (str, None),

        }

    @property
    def description(self) -> str:
        """

        """
        return self._properties.get('description')

    @description.setter
    def description(self, value: str):
        """

        """
        self._properties['description'] = value

    @property
    def url(self) -> str:
        """

        """
        return self._properties.get('url')

    @url.setter
    def url(self, value: str):
        """

        """
        self._properties['url'] = value

    def __enter__(self) -> 'ExternalDocumentation':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class JSON(_kuber_definitions.Definition):
    """
    JSON represents any valid JSON value. These types are
    supported: bool, int64, float64, string, []interface{},
    map[string]interface{} and nil.
    """

    def __init__(
            self,
    ):
        """Create JSON instance."""
        super(JSON, self).__init__(
            api_version='apiextensions/v1beta1',
            kind='JSON'
        )
        self._properties = {

        }
        self._types = {

        }

    def __enter__(self) -> 'JSON':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class JSONSchemaProps(_kuber_definitions.Definition):
    """
    JSONSchemaProps is a JSON-Schema following Specification
    Draft 4 (http://json-schema.org/).
    """

    def __init__(
            self,
            additional_items: 'JSONSchemaPropsOrBool' = None,
            additional_properties: 'JSONSchemaPropsOrBool' = None,
            all_of: typing.List['JSONSchemaProps'] = None,
            any_of: typing.List['JSONSchemaProps'] = None,
            default: 'JSON' = None,
            definitions: dict = None,
            dependencies: dict = None,
            description: str = None,
            enum: typing.List['JSON'] = None,
            example: 'JSON' = None,
            exclusive_maximum: bool = None,
            exclusive_minimum: bool = None,
            external_docs: 'ExternalDocumentation' = None,
            format_: str = None,
            id_: str = None,
            items: 'JSONSchemaPropsOrArray' = None,
            max_items: int = None,
            max_length: int = None,
            max_properties: int = None,
            maximum: float = None,
            min_items: int = None,
            min_length: int = None,
            min_properties: int = None,
            minimum: float = None,
            multiple_of: float = None,
            not_: typing.Optional['JSONSchemaProps'] = None,
            nullable: bool = None,
            one_of: typing.List['JSONSchemaProps'] = None,
            pattern: str = None,
            pattern_properties: dict = None,
            properties: dict = None,
            required: typing.List[str] = None,
            title: str = None,
            type_: str = None,
            unique_items: bool = None,
    ):
        """Create JSONSchemaProps instance."""
        super(JSONSchemaProps, self).__init__(
            api_version='apiextensions/v1beta1',
            kind='JSONSchemaProps'
        )
        self._properties = {
            'additionalItems': additional_items if additional_items is not None else JSONSchemaPropsOrBool(),
            'additionalProperties': additional_properties if additional_properties is not None else JSONSchemaPropsOrBool(),
            'allOf': all_of if all_of is not None else [],
            'anyOf': any_of if any_of is not None else [],
            'default': default if default is not None else JSON(),
            'definitions': definitions if definitions is not None else {},
            'dependencies': dependencies if dependencies is not None else {},
            'description': description if description is not None else '',
            'enum': enum if enum is not None else [],
            'example': example if example is not None else JSON(),
            'exclusiveMaximum': exclusive_maximum if exclusive_maximum is not None else None,
            'exclusiveMinimum': exclusive_minimum if exclusive_minimum is not None else None,
            'externalDocs': external_docs if external_docs is not None else ExternalDocumentation(),
            'format': format_ if format_ is not None else '',
            'id': id_ if id_ is not None else '',
            'items': items if items is not None else JSONSchemaPropsOrArray(),
            'maxItems': max_items if max_items is not None else None,
            'maxLength': max_length if max_length is not None else None,
            'maxProperties': max_properties if max_properties is not None else None,
            'maximum': maximum if maximum is not None else None,
            'minItems': min_items if min_items is not None else None,
            'minLength': min_length if min_length is not None else None,
            'minProperties': min_properties if min_properties is not None else None,
            'minimum': minimum if minimum is not None else None,
            'multipleOf': multiple_of if multiple_of is not None else None,
            'not': not_ if not_ is not None else None,
            'nullable': nullable if nullable is not None else None,
            'oneOf': one_of if one_of is not None else [],
            'pattern': pattern if pattern is not None else '',
            'patternProperties': pattern_properties if pattern_properties is not None else {},
            'properties': properties if properties is not None else {},
            'required': required if required is not None else [],
            'title': title if title is not None else '',
            'type': type_ if type_ is not None else '',
            'uniqueItems': unique_items if unique_items is not None else None,

        }
        self._types = {
            'additionalItems': (JSONSchemaPropsOrBool, None),
            'additionalProperties': (JSONSchemaPropsOrBool, None),
            'allOf': (list, JSONSchemaProps),
            'anyOf': (list, JSONSchemaProps),
            'default': (JSON, None),
            'definitions': (dict, None),
            'dependencies': (dict, None),
            'description': (str, None),
            'enum': (list, JSON),
            'example': (JSON, None),
            'exclusiveMaximum': (bool, None),
            'exclusiveMinimum': (bool, None),
            'externalDocs': (ExternalDocumentation, None),
            'format': (str, None),
            'id': (str, None),
            'items': (JSONSchemaPropsOrArray, None),
            'maxItems': (int, None),
            'maxLength': (int, None),
            'maxProperties': (int, None),
            'maximum': (float, None),
            'minItems': (int, None),
            'minLength': (int, None),
            'minProperties': (int, None),
            'minimum': (float, None),
            'multipleOf': (float, None),
            'not': (JSONSchemaProps, None),
            'nullable': (bool, None),
            'oneOf': (list, JSONSchemaProps),
            'pattern': (str, None),
            'patternProperties': (dict, None),
            'properties': (dict, None),
            'required': (list, str),
            'title': (str, None),
            'type': (str, None),
            'uniqueItems': (bool, None),

        }

    @property
    def additional_items(self) -> 'JSONSchemaPropsOrBool':
        """

        """
        return self._properties.get('additionalItems')

    @additional_items.setter
    def additional_items(self, value: typing.Union['JSONSchemaPropsOrBool', dict]):
        """

        """
        if isinstance(value, dict):
            value = JSONSchemaPropsOrBool().from_dict(value)
        self._properties['additionalItems'] = value

    @property
    def additional_properties(self) -> 'JSONSchemaPropsOrBool':
        """

        """
        return self._properties.get('additionalProperties')

    @additional_properties.setter
    def additional_properties(self, value: typing.Union['JSONSchemaPropsOrBool', dict]):
        """

        """
        if isinstance(value, dict):
            value = JSONSchemaPropsOrBool().from_dict(value)
        self._properties['additionalProperties'] = value

    @property
    def all_of(self) -> typing.List['JSONSchemaProps']:
        """

        """
        return self._properties.get('allOf')

    @all_of.setter
    def all_of(
            self,
            value: typing.Union[typing.List['JSONSchemaProps'], typing.List[dict]]
    ):
        """

        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = JSONSchemaProps().from_dict(item)
            cleaned.append(item)
        self._properties['allOf'] = cleaned

    @property
    def any_of(self) -> typing.List['JSONSchemaProps']:
        """

        """
        return self._properties.get('anyOf')

    @any_of.setter
    def any_of(
            self,
            value: typing.Union[typing.List['JSONSchemaProps'], typing.List[dict]]
    ):
        """

        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = JSONSchemaProps().from_dict(item)
            cleaned.append(item)
        self._properties['anyOf'] = cleaned

    @property
    def default(self) -> 'JSON':
        """

        """
        return self._properties.get('default')

    @default.setter
    def default(self, value: typing.Union['JSON', dict]):
        """

        """
        if isinstance(value, dict):
            value = JSON().from_dict(value)
        self._properties['default'] = value

    @property
    def definitions(self) -> dict:
        """

        """
        return self._properties.get('definitions')

    @definitions.setter
    def definitions(self, value: dict):
        """

        """
        self._properties['definitions'] = value

    @property
    def dependencies(self) -> dict:
        """

        """
        return self._properties.get('dependencies')

    @dependencies.setter
    def dependencies(self, value: dict):
        """

        """
        self._properties['dependencies'] = value

    @property
    def description(self) -> str:
        """

        """
        return self._properties.get('description')

    @description.setter
    def description(self, value: str):
        """

        """
        self._properties['description'] = value

    @property
    def enum(self) -> typing.List['JSON']:
        """

        """
        return self._properties.get('enum')

    @enum.setter
    def enum(
            self,
            value: typing.Union[typing.List['JSON'], typing.List[dict]]
    ):
        """

        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = JSON().from_dict(item)
            cleaned.append(item)
        self._properties['enum'] = cleaned

    @property
    def example(self) -> 'JSON':
        """

        """
        return self._properties.get('example')

    @example.setter
    def example(self, value: typing.Union['JSON', dict]):
        """

        """
        if isinstance(value, dict):
            value = JSON().from_dict(value)
        self._properties['example'] = value

    @property
    def exclusive_maximum(self) -> bool:
        """

        """
        return self._properties.get('exclusiveMaximum')

    @exclusive_maximum.setter
    def exclusive_maximum(self, value: bool):
        """

        """
        self._properties['exclusiveMaximum'] = value

    @property
    def exclusive_minimum(self) -> bool:
        """

        """
        return self._properties.get('exclusiveMinimum')

    @exclusive_minimum.setter
    def exclusive_minimum(self, value: bool):
        """

        """
        self._properties['exclusiveMinimum'] = value

    @property
    def external_docs(self) -> 'ExternalDocumentation':
        """

        """
        return self._properties.get('externalDocs')

    @external_docs.setter
    def external_docs(self, value: typing.Union['ExternalDocumentation', dict]):
        """

        """
        if isinstance(value, dict):
            value = ExternalDocumentation().from_dict(value)
        self._properties['externalDocs'] = value

    @property
    def format_(self) -> str:
        """

        """
        return self._properties.get('format')

    @format_.setter
    def format_(self, value: str):
        """

        """
        self._properties['format'] = value

    @property
    def id_(self) -> str:
        """

        """
        return self._properties.get('id')

    @id_.setter
    def id_(self, value: str):
        """

        """
        self._properties['id'] = value

    @property
    def items(self) -> 'JSONSchemaPropsOrArray':
        """

        """
        return self._properties.get('items')

    @items.setter
    def items(self, value: typing.Union['JSONSchemaPropsOrArray', dict]):
        """

        """
        if isinstance(value, dict):
            value = JSONSchemaPropsOrArray().from_dict(value)
        self._properties['items'] = value

    @property
    def max_items(self) -> int:
        """

        """
        return self._properties.get('maxItems')

    @max_items.setter
    def max_items(self, value: int):
        """

        """
        self._properties['maxItems'] = value

    @property
    def max_length(self) -> int:
        """

        """
        return self._properties.get('maxLength')

    @max_length.setter
    def max_length(self, value: int):
        """

        """
        self._properties['maxLength'] = value

    @property
    def max_properties(self) -> int:
        """

        """
        return self._properties.get('maxProperties')

    @max_properties.setter
    def max_properties(self, value: int):
        """

        """
        self._properties['maxProperties'] = value

    @property
    def maximum(self) -> float:
        """

        """
        return self._properties.get('maximum')

    @maximum.setter
    def maximum(self, value: float):
        """

        """
        self._properties['maximum'] = value

    @property
    def min_items(self) -> int:
        """

        """
        return self._properties.get('minItems')

    @min_items.setter
    def min_items(self, value: int):
        """

        """
        self._properties['minItems'] = value

    @property
    def min_length(self) -> int:
        """

        """
        return self._properties.get('minLength')

    @min_length.setter
    def min_length(self, value: int):
        """

        """
        self._properties['minLength'] = value

    @property
    def min_properties(self) -> int:
        """

        """
        return self._properties.get('minProperties')

    @min_properties.setter
    def min_properties(self, value: int):
        """

        """
        self._properties['minProperties'] = value

    @property
    def minimum(self) -> float:
        """

        """
        return self._properties.get('minimum')

    @minimum.setter
    def minimum(self, value: float):
        """

        """
        self._properties['minimum'] = value

    @property
    def multiple_of(self) -> float:
        """

        """
        return self._properties.get('multipleOf')

    @multiple_of.setter
    def multiple_of(self, value: float):
        """

        """
        self._properties['multipleOf'] = value

    @property
    def not_(self) -> typing.Optional['JSONSchemaProps']:
        """

        """
        return self._properties.get('not')

    @not_.setter
    def not_(self, value: typing.Union['JSONSchemaProps', dict, None]):
        """

        """
        if isinstance(value, dict):
            value = JSONSchemaProps().from_dict(value)
        self._properties['not'] = value

    @property
    def nullable(self) -> bool:
        """

        """
        return self._properties.get('nullable')

    @nullable.setter
    def nullable(self, value: bool):
        """

        """
        self._properties['nullable'] = value

    @property
    def one_of(self) -> typing.List['JSONSchemaProps']:
        """

        """
        return self._properties.get('oneOf')

    @one_of.setter
    def one_of(
            self,
            value: typing.Union[typing.List['JSONSchemaProps'], typing.List[dict]]
    ):
        """

        """
        cleaned = []
        for item in value:
            if isinstance(item, dict):
                item = JSONSchemaProps().from_dict(item)
            cleaned.append(item)
        self._properties['oneOf'] = cleaned

    @property
    def pattern(self) -> str:
        """

        """
        return self._properties.get('pattern')

    @pattern.setter
    def pattern(self, value: str):
        """

        """
        self._properties['pattern'] = value

    @property
    def pattern_properties(self) -> dict:
        """

        """
        return self._properties.get('patternProperties')

    @pattern_properties.setter
    def pattern_properties(self, value: dict):
        """

        """
        self._properties['patternProperties'] = value

    @property
    def properties(self) -> dict:
        """

        """
        return self._properties.get('properties')

    @properties.setter
    def properties(self, value: dict):
        """

        """
        self._properties['properties'] = value

    @property
    def required(self) -> typing.List[str]:
        """

        """
        return self._properties.get('required')

    @required.setter
    def required(self, value: typing.List[str]):
        """

        """
        self._properties['required'] = value

    @property
    def title(self) -> str:
        """

        """
        return self._properties.get('title')

    @title.setter
    def title(self, value: str):
        """

        """
        self._properties['title'] = value

    @property
    def type_(self) -> str:
        """

        """
        return self._properties.get('type')

    @type_.setter
    def type_(self, value: str):
        """

        """
        self._properties['type'] = value

    @property
    def unique_items(self) -> bool:
        """

        """
        return self._properties.get('uniqueItems')

    @unique_items.setter
    def unique_items(self, value: bool):
        """

        """
        self._properties['uniqueItems'] = value

    def __enter__(self) -> 'JSONSchemaProps':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class JSONSchemaPropsOrArray(_kuber_definitions.Definition):
    """
    JSONSchemaPropsOrArray represents a value that can either be
    a JSONSchemaProps or an array of JSONSchemaProps. Mainly
    here for serialization purposes.
    """

    def __init__(
            self,
    ):
        """Create JSONSchemaPropsOrArray instance."""
        super(JSONSchemaPropsOrArray, self).__init__(
            api_version='apiextensions/v1beta1',
            kind='JSONSchemaPropsOrArray'
        )
        self._properties = {

        }
        self._types = {

        }

    def __enter__(self) -> 'JSONSchemaPropsOrArray':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class JSONSchemaPropsOrBool(_kuber_definitions.Definition):
    """
    JSONSchemaPropsOrBool represents JSONSchemaProps or a
    boolean value. Defaults to true for the boolean property.
    """

    def __init__(
            self,
    ):
        """Create JSONSchemaPropsOrBool instance."""
        super(JSONSchemaPropsOrBool, self).__init__(
            api_version='apiextensions/v1beta1',
            kind='JSONSchemaPropsOrBool'
        )
        self._properties = {

        }
        self._types = {

        }

    def __enter__(self) -> 'JSONSchemaPropsOrBool':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class JSONSchemaPropsOrStringArray(_kuber_definitions.Definition):
    """
    JSONSchemaPropsOrStringArray represents a JSONSchemaProps or
    a string array.
    """

    def __init__(
            self,
    ):
        """Create JSONSchemaPropsOrStringArray instance."""
        super(JSONSchemaPropsOrStringArray, self).__init__(
            api_version='apiextensions/v1beta1',
            kind='JSONSchemaPropsOrStringArray'
        )
        self._properties = {

        }
        self._types = {

        }

    def __enter__(self) -> 'JSONSchemaPropsOrStringArray':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class ServiceReference(_kuber_definitions.Definition):
    """
    ServiceReference holds a reference to Service.legacy.k8s.io
    """

    def __init__(
            self,
            name: str = None,
            namespace: str = None,
            path: str = None,
    ):
        """Create ServiceReference instance."""
        super(ServiceReference, self).__init__(
            api_version='apiextensions/v1beta1',
            kind='ServiceReference'
        )
        self._properties = {
            'name': name if name is not None else '',
            'namespace': namespace if namespace is not None else '',
            'path': path if path is not None else '',

        }
        self._types = {
            'name': (str, None),
            'namespace': (str, None),
            'path': (str, None),

        }

    @property
    def name(self) -> str:
        """
        `name` is the name of the service. Required
        """
        return self._properties.get('name')

    @name.setter
    def name(self, value: str):
        """
        `name` is the name of the service. Required
        """
        self._properties['name'] = value

    @property
    def namespace(self) -> str:
        """
        `namespace` is the namespace of the service. Required
        """
        return self._properties.get('namespace')

    @namespace.setter
    def namespace(self, value: str):
        """
        `namespace` is the namespace of the service. Required
        """
        self._properties['namespace'] = value

    @property
    def path(self) -> str:
        """
        `path` is an optional URL path which will be sent in any
        request to this service.
        """
        return self._properties.get('path')

    @path.setter
    def path(self, value: str):
        """
        `path` is an optional URL path which will be sent in any
        request to this service.
        """
        self._properties['path'] = value

    def __enter__(self) -> 'ServiceReference':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class WebhookClientConfig(_kuber_definitions.Definition):
    """
    WebhookClientConfig contains the information to make a TLS
    connection with the webhook. It has the same field as
    admissionregistration.v1beta1.WebhookClientConfig.
    """

    def __init__(
            self,
            ca_bundle: str = None,
            service: 'ServiceReference' = None,
            url: str = None,
    ):
        """Create WebhookClientConfig instance."""
        super(WebhookClientConfig, self).__init__(
            api_version='apiextensions/v1beta1',
            kind='WebhookClientConfig'
        )
        self._properties = {
            'caBundle': ca_bundle if ca_bundle is not None else '',
            'service': service if service is not None else ServiceReference(),
            'url': url if url is not None else '',

        }
        self._types = {
            'caBundle': (str, None),
            'service': (ServiceReference, None),
            'url': (str, None),

        }

    @property
    def ca_bundle(self) -> str:
        """
        `caBundle` is a PEM encoded CA bundle which will be used to
        validate the webhook's server certificate. If unspecified,
        system trust roots on the apiserver are used.
        """
        return self._properties.get('caBundle')

    @ca_bundle.setter
    def ca_bundle(self, value: str):
        """
        `caBundle` is a PEM encoded CA bundle which will be used to
        validate the webhook's server certificate. If unspecified,
        system trust roots on the apiserver are used.
        """
        self._properties['caBundle'] = value

    @property
    def service(self) -> 'ServiceReference':
        """
        `service` is a reference to the service for this webhook.
        Either `service` or `url` must be specified.

        If the webhook is running within the cluster, then you
        should use `service`.

        Port 443 will be used if it is open, otherwise it is an
        error.
        """
        return self._properties.get('service')

    @service.setter
    def service(self, value: typing.Union['ServiceReference', dict]):
        """
        `service` is a reference to the service for this webhook.
        Either `service` or `url` must be specified.

        If the webhook is running within the cluster, then you
        should use `service`.

        Port 443 will be used if it is open, otherwise it is an
        error.
        """
        if isinstance(value, dict):
            value = ServiceReference().from_dict(value)
        self._properties['service'] = value

    @property
    def url(self) -> str:
        """
        `url` gives the location of the webhook, in standard URL
        form (`scheme://host:port/path`). Exactly one of `url` or
        `service` must be specified.

        The `host` should not refer to a service running in the
        cluster; use the `service` field instead. The host might be
        resolved via external DNS in some apiservers (e.g., `kube-
        apiserver` cannot resolve in-cluster DNS as that would be a
        layering violation). `host` may also be an IP address.

        Please note that using `localhost` or `127.0.0.1` as a
        `host` is risky unless you take great care to run this
        webhook on all hosts which run an apiserver which might need
        to make calls to this webhook. Such installs are likely to
        be non-portable, i.e., not easy to turn up in a new cluster.

        The scheme must be "https"; the URL must begin with
        "https://".

        A path is optional, and if present may be any string
        permissible in a URL. You may use the path to pass an
        arbitrary string to the webhook, for example, a cluster
        identifier.

        Attempting to use a user or basic auth e.g. "user:password@"
        is not allowed. Fragments ("#...") and query parameters
        ("?...") are not allowed, either.
        """
        return self._properties.get('url')

    @url.setter
    def url(self, value: str):
        """
        `url` gives the location of the webhook, in standard URL
        form (`scheme://host:port/path`). Exactly one of `url` or
        `service` must be specified.

        The `host` should not refer to a service running in the
        cluster; use the `service` field instead. The host might be
        resolved via external DNS in some apiservers (e.g., `kube-
        apiserver` cannot resolve in-cluster DNS as that would be a
        layering violation). `host` may also be an IP address.

        Please note that using `localhost` or `127.0.0.1` as a
        `host` is risky unless you take great care to run this
        webhook on all hosts which run an apiserver which might need
        to make calls to this webhook. Such installs are likely to
        be non-portable, i.e., not easy to turn up in a new cluster.

        The scheme must be "https"; the URL must begin with
        "https://".

        A path is optional, and if present may be any string
        permissible in a URL. You may use the path to pass an
        arbitrary string to the webhook, for example, a cluster
        identifier.

        Attempting to use a user or basic auth e.g. "user:password@"
        is not allowed. Fragments ("#...") and query parameters
        ("?...") are not allowed, either.
        """
        self._properties['url'] = value

    def __enter__(self) -> 'WebhookClientConfig':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False



from pytest import mark

from kuber.latest import apps_v1
from kuber.latest import batch_v1
from kuber.latest import core_v1

RESOURCES_TO_TEST = [core_v1.Pod, apps_v1.Deployment, batch_v1.Job]


@mark.parametrize('resource_class', RESOURCES_TO_TEST)
def test_container_methods(resource_class):
    """Tests auto-generated container methods."""
    resource = resource_class()
    resource.append_container(name='foo', image='foo')
    resource.append_container(name='bar', image='bar')
    assert resource.get_container('foo').name == 'foo'
    assert resource.get_container('bar').name == 'bar'
    assert len(resource.get_containers()) == 2
